/*
 * SURF is a freeware internet surfing tool to use in Windows 95.
 * It's a list driven web agent, fetching only pages you request.
 * It might best be categorized as a web page harvesting program.
 * Content analysis helps you deal with the flood of information.
 *
 * Copyright (C) 1998 Glenn Scheper. This program is free software;
 * you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details. You should have
 * received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Contact SURF's author Glenn Scheper at <URL:mailto:scheper@hughes.net>.
 * Download SURF.CPP and SURF.EXE at <URL:http://hughes.net/~scheper/>.
 * GNU General Public License: <URL:http://www.fsf.org/copyleft/gpl.html>.
 *
 * Compiled under Microsoft Developer Studio and
 * Microsoft Visual C++ 5.0 compiler as follows:
 * 1.) Copy surf.cpp to a new directory surf.
 * 2.) File, Open, surf.cpp.
 * 3.) Build, Rebuild All: Create default project? Yes.
 * 4.) Build, Set Active Configuration: Release.
 * 5.) Project, Settings: C/C++, Code Generation: Multithreaded.
 * Also add something to Project Settings: (I used 200: /Zm200)
 * to prevent "fatal error C1076: compiler limit : internal heap
 * limit reached; use /Zm to specify a higher limit"
 * 6.) Build, Rebuild all.
 * 7.) Copy ./release/surf.exe to a directory in the path.
 */

char SurfVersion [] = "SURF v1.3"; //For the http agent line
char SurfVersionAndDate [] = "SURF version 1.3, April 12, 1998.";

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <conio.h>
#include <ctype.h>
#include <direct.h>
#include <afx.h>
#include <afxwin.h>
#include <afxinet.h>
#include <iostream.h>

void GiveUsageMessage()
{
	fprintf(stderr,
"Surf fetches pages using http, parses and analyzes the page contents,\n"
"saves local copies of wrapped text and URLs, and makes a list of URLs.\n"
"\n"
"  SURF LIST   -- Fetches URLs in LIST not marked by * or # next line.\n"
"  SURF LIST -F URL1 ...  -- Fetches URL1, ..., writes to local files.\n"
"  SURF LIST -A FILE* ...  -- Augments list by local FILE* HTML files.\n"
"  SURF LIST -Q QUERY KEYS ...  -- Adds search engine queries to list.\n"
"  SURF LIST -M  -- Moves files according to controls you put in LIST.\n"
"  SURF LIST -W > SURFED.HTM  -- Writes an HTML list of fetched links.\n"
"\n"
"To fetch some, more, or all of the web pages listed in URL file, LIST,\n"
"First manually edit file LIST to delete asterisks on lines immediately\n"
"following lines showing URLs you want to fetch. Then say: 'SURF LIST'.\n"
"\n"
" %s Copyright (C) 1998 Glenn Scheper.\n"   //...SurfVersionAndDate here
" SURF comes with ABSOLUTELY NO WARRANTY; See source code for details.\n"
" This is free software, and you are welcome to redistribute it under\n"
" certain conditions; See source code for details. <scheper@hughes.net>.\n"
" Download SURF.CPP and SURF.EXE from <URL:http://hughes.net/~scheper/>.\n"
" To fetch my home page, say: 'surf list -f http://hughes.net/~scheper'.\n",
    SurfVersionAndDate );
	return;
}

/*
 * History of Glenn Scheper's own versions of SURF:
 * Download latest source and Win-95 executable at:
 * <URL:http://hughes.net/~scheper/surf.cpp>
 * <URL:http://hughes.net/~scheper/surf.exe>
 *
 * Various "PRE-SURF" program versions parsed HTML files, etc.
 *
 * 12-31-97 First reasonably functional SURF.EXE fetches HTML files
 * at random as given in a URL list file, parses the HTML files and
 * saves the text + links parts on local drive, and grooms URL list.
 *
 * 2-23-98 Fixed lack of (LPCTSTR) typecast on a TCHAR named szErr.
 * 2-23-98 Aggressive work adding file vocabulary info to list file.

 * SURF Version 1.2, Mar 22 1998:
 * Started a version number, to fit the GNU licensing requirements.
 * Added GNU style copyright notice to source and to usage message.
 * Totally reorganized command line invocation format with a -FLAG.
 * Fixed random numbers: 32K was too small to span a huge URL list.
 * Changed the sort on cached URLs to use sum of 3 quality indices.
 * Improved the handling of titles and URL notes on AREA and FRAME.
 * Added the URL List -> HTML file and Query URLs -> List features.

 * SURF Version 1.3, April 12 1998:
 * I no longer require a user to type http:// before the url argument.
 * I now output the two lines of most used words when fetching a file.
 * I changed to no-fetch asterisk all but four popular search engines.
 * I fixed the LocalFilename to use \ rather than / as path separator.
 * I fixed so final word of URL annotation is not lost during fetches.
 * I added most-used-of-the-most-used-words line to LIST and LIST.HTM.
 * I added File Naming Control '@path word word..." idea to LIST file.
 * I added File Naming Control and -M option to rearrange local files.
 * I added URL annotation based on title text for AREA and FRAME tags.
 * I changed third quality metric from word count to vocabulary count.

 *
 *
 */

/********************************************************

	///wip... comments mark places for future improvements.

I need to clone stuff for renames after a -F fetch,
but doing a -M after the fetches is its workaround,
and I'd like to get all these new features out now.
So, this next comment is not quite ready yet...
 * I added File Naming Control for -F option to control storing files.
In addition, -F and -M need robust fallback if rename filename2 exists.
I'd like to force first letter of filename a-z based on the QQQ index!

Bug: -A reparse of local files adds new AREA/FRAME text onto word list.
Ensure Area, Frame, etc. write anchor text, to keep reparse the same.

The -Q query is a new feature, needs elaboration!
Query topics need some character escaping to %xx form.
List whether search engines take simple/advanced queries.
tune up the query list: locate more search engines URL;
improve queries; discard the no-good engines;
have categorization as to topic or other specialty.


Add a process to evaluate FORMS and therefore create a query URL.


I should change how to handle files that actually contain BASE tags:
I should list under my base tag, but use theirs to resolve local urls?


This is an entirely different BASE issue:
If parsing local file w/o BASE tag, create a URL such as "file://....",
because otherwise, all local files w/o base tags will merge in one item.


Add a stop list (or user could copy to init list).


Process and record redirections during the fetch:
 3.3 An important observation of RFC1808 is that
 the BASE href must come from final redirection.
 That will influence how I perform GetURL stuff.


Later, do not compromise on the URL parse, by a too-early test
for no locn && no pathpart in the routine CombineLinkUrlWithBase():
I've seen pages with BASE=query, Anchors=new query part only.


Add a new concept to re-extract small files from a COPY *.HTM wad.
E.g., add a search for my added @end comment, to reset per-file parsing.
If only surf wrote <base>, every BASE could trigger a top-of-file reset.


I'm never happy building upon obscure library code blocks,
so I'd like to plumb the meaning of the URL fetching work, at
comment: Adapt my loop to the stream read of the sample code.


Someday, support ftp: and gopher: methods.
Until then, I could at least permit them to stay in the list?

********************************************************/


	//This lookup table operates like SortedEntityNames,
	//listing words to ignore of 3-15 chars + one space.
	//This string is first to not overtax compiler heap.

	//I'd appreciate feedback to better adjust this list:
	//(and for anything else you'd like in this program!)
	//E-mail me: Glenn Scheper - scheper@hughes.net

char CommonWordList [] = {
"! .............."  //low sentinel
"aber ..........."
"able ..........."
"about .........."
"above .........."
"absolutely ....."
"accept ........."
"accepted ......."
"access ........."
"according ......"
"account ........"
"accounts ......."
"achieved ......."
"acrobat ........"
"across ........."
"act ............"
"activities ....."
"activity ......."
"acts ..........."
"actual ........."
"actually ......."
"add ............"
"added .........."
"adding ........."
"addition ......."
"additional ....."
"additions ......"
"address ........"
"addressed ......"
"addresses ......"
"adds ..........."
"ads ............"
"after .........."
"afterwards ....."
"again .........."
"against ........"
"age ............"
"ago ............"
"ahead .........."
"aim ............"
"air ............"
"aka ............"
"all ............"
"allow .........."
"allows ........."
"almost ........."
"alone .........."
"along .........."
"already ........"
"als ............"
"also ..........."
"alt ............"
"altered ........"
"although ......."
"always ........."
"ammendment ....."
"among .........."
"amount ........."
"and ............"
"announced ......"
"another ........"
"answer ........."
"answered ......."
"answers ........"
"anti ..........."
"any ............"
"anyone ........."
"anything ......."
"anyway ........."
"anywhere ......."
"aol ............"
"apart .........."
"apparent ......."
"apparently ....."
"appear ........."
"appearance ....."
"appeared ......."
"appears ........"
"applications ..."
"applied ........"
"apply .........."
"approach ......."
"approximately .."
"apr ............"
"april .........."
"archive ........"
"archives ......."
"are ............"
"area ..........."
"areas .........."
"aren ..........."
"around ........."
"article ........"
"articles ......."
"aside .........."
"ask ............"
"asked .........."
"asking ........."
"asp ............"
"aspect ........."
"aspects ........"
"associates ....."
"assume ........."
"assumed ........"
"assumes ........"
"attempt ........"
"attempted ......"
"attempts ......."
"auch ..........."
"audio .........."
"auf ............"
"aug ............"
"august ........."
"aus ............"
"author ........."
"authors ........"
"available ......"
"avait .........."
"ave ............"
"avoid .........."
"award .........."
"awards ........."
"aware .........."
"away ..........."
"back ..........."
"background ....."
"bad ............"
"banner ........."
"base ..........."
"based .........."
"basic .........."
"basically ......"
"basis .........."
"bbs ............"
"bear ..........."
"became ........."
"because ........"
"become ........."
"becomes ........"
"becoming ......."
"been ..........."
"before ........."
"began .........."
"begin .........."
"beginning ......"
"begun .........."
"behind ........."
"behold ........."
"bei ............"
"being .........."
"believe ........"
"belong ........."
"below .........."
"beside ........."
"besides ........"
"best ..........."
"better ........."
"between ........"
"beyond ........."
"big ............"
"bin ............"
"bit ............"
"book ..........."
"bookmark ......."
"bookmarks ......"
"books .........."
"both ..........."
"bottom ........."
"box ............"
"brain .........."
"bring .........."
"bringing ......."
"brings ........."
"broad .........."
"brought ........"
"browse ........."
"browser ........"
"build .........."
"built .........."
"business ......."
"but ............"
"button ........."
"byte ..........."
"bytes .........."
"call ..........."
"called ........."
"calling ........"
"calls .........."
"cam ............"
"came ..........."
"can ............"
"cannot ........."
"carried ........"
"carry .........."
"case ..........."
"catalog ........"
"caught ........."
"cause .........."
"caused ........."
"causes ........."
"causing ........"
"cease .........."
"ceased ........."
"ceases ........."
"center ........."
"central ........"
"certain ........"
"certainly ......"
"cgi ............"
"chance ........."
"change ........."
"changed ........"
"changes ........"
"changing ......."
"chapter ........"
"chapters ......."
"character ......"
"characters ....."
"charge ........."
"chat ..........."
"check .........."
"choice ........."
"choose ........."
"chosen ........."
"circa .........."
"clean .........."
"clear .........."
"clearly ........"
"click .........."
"close .........."
"closely ........"
"closer ........."
"com ............"
"come ..........."
"comes .........."
"coming ........."
"command ........"
"comment ........"
"comments ......."
"common ........."
"commonly ......."
"comparative ...."
"compare ........"
"compared ......."
"complete ......."
"completed ......"
"completely ....."
"complex ........"
"comprehensive .."
"compuserve ....."
"computer ......."
"concept ........"
"concepts ......."
"concern ........"
"concerned ......"
"concerning ....."
"concerns ......."
"conclusions ...."
"connect ........"
"connection ....."
"connections ...."
"consider ......."
"considered ....."
"considers ......"
"consisting ....."
"consists ......."
"constant ......."
"constantly ....."
"contact ........"
"contacts ......."
"contain ........"
"contained ......"
"containing ....."
"contains ......."
"contents ......."
"context ........"
"continue ......."
"continued ......"
"continues ......"
"control ........"
"cool ..........."
"copy ..........."
"copyright ......"
"cor ............"
"correspondence ."
"corresponding .."
"could .........."
"couldn ........."
"count .........."
"counter ........"
"course ........."
"cover .........."
"covered ........"
"create ........."
"created ........"
"creating ......."
"creator ........"
"current ........"
"currently ......"
"cut ............"
"dan ............"
"dans ..........."
"das ............"
"data ..........."
"date ..........."
"dates .........."
"day ............"
"days ..........."
"dealing ........"
"death .........."
"dec ............"
"december ......."
"declared ......."
"deep ..........."
"default ........"
"define ........."
"defined ........"
"definitely ....."
"degree ........."
"del ............"
"dem ............"
"den ............"
"department ....."
"depend ........."
"depending ......"
"depends ........"
"dept ..........."
"depth .........."
"der ............"
"des ............"
"describe ......."
"described ......"
"describes ......"
"description ...."
"descriptions ..."
"design ........."
"designed ......."
"desire ........."
"despite ........"
"detail ........."
"detailed ......."
"details ........"
"developed ......"
"developers ....."
"developing ....."
"development ...."
"device ........."
"did ............"
"didn ..........."
"die ............"
"dies ..........."
"diese .........."
"difference ....."
"differences ...."
"different ......"
"differently ...."
"difficult ......"
"digital ........"
"dir ............"
"direct ........."
"directly ......."
"directory ......"
"disclaimer ....."
"discover ......."
"discovery ......"
"discuss ........"
"discussion ....."
"display ........"
"diverse ........"
"docs ..........."
"document ......."
"documents ......"
"does ..........."
"doesn .........."
"doing .........."
"don ............"
"done ..........."
"doubt .........."
"down ..........."
"download ......."
"draw ..........."
"due ............"
"durch .........."
"during ........."
"dynamic ........"
"each ..........."
"earlier ........"
"early .........."
"easier ........."
"easily ........."
"easy ..........."
"eat ............"
"edited ........."
"edition ........"
"editor ........."
"editors ........"
"edu ............"
"effect ........."
"effects ........"
"eight .........."
"eighth ........."
"ein ............"
"eine ..........."
"einem .........."
"einer .........."
"either ........."
"electronic ....."
"elements ......."
"eleven ........."
"eleventh ......."
"else ..........."
"elsewhere ......"
"email .........."
"encounter ......"
"end ............"
"ended .........."
"ending ........."
"ends ..........."
"english ........"
"enhanced ......."
"enough ........."
"enter .........."
"entered ........"
"entering ......."
"entire ........."
"entirely ......."
"entry .........."
"equal .........."
"equally ........"
"errors ........."
"especially ....."
"essence ........"
"essentially ...."
"etaient ........"
"etait .........."
"etc ............"
"even ..........."
"event .........."
"events ........."
"eventually ....."
"ever ..........."
"every .........."
"everyone ......."
"everything ....."
"everywhere ....."
"exact .........."
"exactly ........"
"examine ........"
"example ........"
"examples ......."
"excellent ......"
"except ........."
"excepted ......."
"exception ......"
"exchange ......."
"exe ............"
"exist .........."
"existed ........"
"existence ......"
"existing ......."
"exists ........."
"expansion ......"
"expect ........."
"expected ......."
"expects ........"
"explain ........"
"explanation ...."
"explorer ......."
"extend ........."
"extending ......"
"extends ........"
"extension ......"
"extensive ......"
"extra .........."
"extreme ........"
"extremely ......"
"fact ..........."
"facts .........."
"fair ..........."
"faith .........."
"fall ..........."
"fallen ........."
"falling ........"
"falls .........."
"false .........."
"familiar ......."
"family ........."
"far ............"
"fast ..........."
"feature ........"
"features ......."
"feb ............"
"february ......."
"feedback ......."
"feel ..........."
"fell ..........."
"few ............"
"field .........."
"fields ........."
"fifth .........."
"fifty .........."
"figure ........."
"file ..........."
"files .........."
"filled ........."
"final .........."
"finally ........"
"find ..........."
"finding ........"
"finds .........."
"finish ........."
"first .........."
"fit ............"
"five ..........."
"follow ........."
"followed ......."
"following ......"
"follows ........"
"for ............"
"forever ........"
"form ..........."
"format ........."
"formats ........"
"formed ........."
"former ........."
"forms .........."
"forth .........."
"forum .........."
"forward ........"
"found .........."
"four ..........."
"fourth ........."
"frame .........."
"frames ........."
"free ..........."
"freedom ........"
"frequent ......."
"frequently ....."
"fri ............"
"friday ........."
"from ..........."
"front .........."
"ftp ............"
"fulfilled ......"
"full ..........."
"fully .........."
"fur ............"
"further ........"
"future ........."
"gallery ........"
"gateway ........"
"gathered ......."
"gave ..........."
"general ........"
"generally ......"
"geocities ......"
"get ............"
"gets ..........."
"getting ........"
"gif ............"
"give ..........."
"given .........."
"gives .........."
"goes ..........."
"going .........."
"gold ..........."
"gone ..........."
"good ..........."
"got ............"
"gov ............"
"graphics ......."
"great .........."
"greater ........"
"greatly ........"
"group .........."
"guestbook ......"
"guide .........."
"haben .........."
"had ............"
"half ..........."
"happen ........."
"happened ......."
"happening ......"
"hard ..........."
"has ............"
"hast ..........."
"hath ..........."
"have ..........."
"haven .........."
"having ........."
"hear ..........."
"heard .........."
"hears .........."
"held ..........."
"help ..........."
"helps .........."
"hence .........."
"her ............"
"here ..........."
"herein ........."
"herself ........"
"high ..........."
"higher ........."
"highest ........"
"highly ........."
"him ............"
"himself ........"
"his ............"
"hold ..........."
"holding ........"
"holds .........."
"home ..........."
"homepage ......."
"homepages ......"
"hope ..........."
"host ..........."
"hot ............"
"hour ..........."
"hours .........."
"how ............"
"however ........"
"href ..........."
"htm ............"
"html ..........."
"http ..........."
"huge ..........."
"hundred ........"
"hundreds ......."
"hundredth ......"
"idea ..........."
"ideal .........."
"ideas .........."
"identity ......."
"ignored ........"
"ihm ............"
"ils ............"
"image .........."
"images ........."
"imagined ......."
"immediate ......"
"immediately ...."
"implied ........"
"imply .........."
"importance ....."
"important ......"
"improve ........"
"inc ............"
"include ........"
"included ......."
"includes ......."
"including ......"
"increase ......."
"increasing ....."
"indeed ........."
"independent ...."
"index .........."
"indicate ......."
"indicated ......"
"infer .........."
"info ..........."
"information ...."
"infoseek ......."
"initial ........"
"inner .........."
"input .........."
"insert ........."
"inside ........."
"insight ........"
"install ........"
"instance ......."
"instant ........"
"instead ........"
"institute ......"
"intended ......."
"inter .........."
"interactive ...."
"interest ......."
"interested ....."
"interesting ...."
"interests ......"
"internal ......."
"international .."
"internet ......."
"into ..........."
"introduction ..."
"invoke ........."
"involved ......."
"involves ......."
"irc ............"
"island ........."
"issue .........."
"issues ........."
"ist ............"
"item ..........."
"items .........."
"its ............"
"itself ........."
"jan ............"
"january ........"
"java ..........."
"join ..........."
"journals ......."
"jpg ............"
"jul ............"
"july ..........."
"jun ............"
"june ..........."
"just ..........."
"keep ..........."
"keeping ........"
"keeps .........."
"kept ..........."
"key ............"
"keyboard ......."
"kind ..........."
"kinds .........."
"king ..........."
"knew ..........."
"know ..........."
"knowing ........"
"knowledge ......"
"known .........."
"knows .........."
"lack ..........."
"language ......."
"languages ......"
"large .........."
"largely ........"
"las ............"
"last ..........."
"lasted ........."
"lastly ........."
"late ..........."
"lately ........."
"later .........."
"latest ........."
"latter ........."
"lay ............"
"lead ..........."
"leads .........."
"learn .........."
"least .........."
"leave .........."
"leaving ........"
"led ............"
"left ..........."
"length ........."
"les ............"
"less ..........."
"lesser ........."
"lest ..........."
"let ............"
"letter ........."
"letters ........"
"level .........."
"levels ........."
"lib ............"
"liber .........."
"library ........"
"license ........"
"lie ............"
"lies ..........."
"life ..........."
"lifted ........."
"light .........."
"like ..........."
"liked .........."
"likely ........."
"likened ........"
"likes .........."
"liking ........."
"limited ........"
"line ..........."
"lines .........."
"link ..........."
"linked ........."
"links .........."
"list ..........."
"listed ........."
"listen ........."
"lists .........."
"literal ........"
"literally ......"
"little ........."
"live ..........."
"lived .........."
"living ........."
"local .........."
"located ........"
"location ......."
"long ..........."
"longer ........."
"look ..........."
"looked ........."
"looking ........"
"looks .........."
"los ............"
"lot ............"
"lots ..........."
"low ............"
"lower .........."
"lui ............"
"machine ........"
"made ..........."
"mail ..........."
"mailing ........"
"mailto ........."
"main ..........."
"mainly ........."
"mais ..........."
"major .........."
"make ..........."
"makes .........."
"making ........."
"man ............"
"manner ........."
"many ..........."
"map ............"
"mar ............"
"march .........."
"mark ..........."
"marked ........."
"marker ........."
"marks .........."
"matter ........."
"matters ........"
"may ............"
"maybe .........."
"mean ..........."
"meaning ........"
"means .........."
"meant .........."
"med ............"
"meet ..........."
"meeting ........"
"meetings ......."
"meets .........."
"member ........."
"members ........"
"men ............"
"mentioned ......"
"menu ..........."
"menus .........."
"mere ..........."
"merely ........."
"message ........"
"messages ......."
"meta ..........."
"method ........."
"methods ........"
"microsoft ......"
"mid ............"
"midst .........."
"might .........."
"mind ..........."
"mine ..........."
"minute ........."
"minutes ........"
"misc ..........."
"miscellaneous .."
"missing ........"
"mission ........"
"mixed .........."
"moment ........."
"mon ............"
"monday ........."
"monitor ........"
"month .........."
"months ........."
"more ..........."
"moreover ......."
"most ..........."
"mostly ........."
"moved .........."
"movement ......."
"moving ........."
"msg ............"
"much ..........."
"multi .........."
"must ..........."
"myself ........."
"nach ..........."
"name ..........."
"named .........."
"namely ........."
"names .........."
"nations ........"
"naturally ......"
"navigator ......"
"near ..........."
"neat ..........."
"neatly ........."
"necessarily ...."
"necessary ......"
"need ..........."
"needed ........."
"needing ........"
"needs .........."
"neither ........"
"net ............"
"netscape ......."
"network ........"
"never .........."
"new ............"
"news ..........."
"newsletter ....."
"next ..........."
"nice ..........."
"nicely ........."
"nicht .........."
"night .........."
"nine ..........."
"nineth ........."
"ninety ........."
"ninth .........."
"nobody ........."
"non ............"
"none ..........."
"noone .........."
"nor ............"
"normal ........."
"normally ......."
"not ............"
"note ..........."
"noted .........."
"notes .........."
"nothing ........"
"notice ........."
"noting ........."
"nov ............"
"november ......."
"now ............"
"nowhere ........"
"number ........."
"nur ............"
"object ........."
"objects ........"
"obtain ........."
"obviously ......"
"occur .........."
"occurred ......."
"occurs ........."
"oct ............"
"october ........"
"oder ..........."
"off ............"
"offer .........."
"offered ........"
"offering ......."
"offers ........."
"office ........."
"often .........."
"old ............"
"once ..........."
"one ............"
"ones ..........."
"online ........."
"only ..........."
"onto ..........."
"open ..........."
"opened ........."
"opposite ......."
"order .........."
"orders ........."
"ordinary ......."
"org ............"
"original ......."
"originally ....."
"other .........."
"others ........."
"otherwise ......"
"ought .........."
"our ............"
"ours ..........."
"ourselves ......"
"out ............"
"outer .........."
"outside ........"
"over ..........."
"overly ........."
"overview ......."
"owed ..........."
"owes ..........."
"owing .........."
"own ............"
"page ..........."
"pages .........."
"pair ..........."
"paired ........."
"papers ........."
"part ..........."
"particular ....."
"particularly ..."
"partly ........."
"parts .........."
"pass ..........."
"passage ........"
"passed ........."
"past ..........."
"path ..........."
"patient ........"
"pdf ............"
"per ............"
"perfect ........"
"perfectly ......"
"perhaps ........"
"period ........."
"permission ....."
"person ........."
"personal ......."
"personality ...."
"perspective ...."
"pictures ......."
"pieces ........."
"place .........."
"placed ........."
"places ........."
"plain .........."
"plans .........."
"please ........."
"pleasure ......."
"plus ..........."
"point .........."
"points ........."
"policy ........."
"possess ........"
"possibilities .."
"possibility ...."
"possible ......."
"possibly ......."
"post ..........."
"posted ........."
"potential ......"
"pour ..........."
"powerful ......."
"practical ......"
"practice ......."
"pre ............"
"precedes ......."
"preceding ......"
"preface ........"
"prefer ........."
"prepared ......."
"present ........"
"presented ......"
"presents ......."
"preserve ......."
"press .........."
"pretty ........."
"prev ..........."
"previous ......."
"previously ....."
"primarily ......"
"primary ........"
"print .........."
"prior .........."
"pro ............"
"probable ......."
"probably ......."
"problem ........"
"problems ......."
"proceed ........"
"process ........"
"processes ......"
"produce ........"
"produced ......."
"product ........"
"program ........"
"programs ......."
"project ........"
"proper ........."
"provide ........"
"provided ......."
"provides ......."
"providing ......"
"pub ............"
"publication ...."
"publications ..."
"published ......"
"publisher ......"
"publishing ....."
"purely ........."
"purpose ........"
"purposes ......."
"put ............"
"puts ..........."
"putting ........"
"quality ........"
"que ............"
"query .........."
"question ......."
"questions ......"
"qui ............"
"quick .........."
"quickly ........"
"quicktime ......"
"quite .........."
"quot ..........."
"quote .........."
"quoted ........."
"quotes ........."
"raise .........."
"raised ........."
"rarely ........."
"rather ........."
"reach .........."
"reached ........"
"read ..........."
"reader ........."
"readily ........"
"reads .........."
"ready .........."
"real ..........."
"reality ........"
"realize ........"
"really ........."
"reason ........."
"reasons ........"
"receive ........"
"received ......."
"recent ........."
"recently ......."
"record ........."
"recorded ......."
"refer .........."
"reference ......"
"references ....."
"referencing ...."
"referring ......"
"refers ........."
"regard ........."
"regarding ......"
"regardless ....."
"register ......."
"regular ........"
"related ........"
"relative ......."
"relatively ....."
"release ........"
"relevant ......."
"remain ........."
"remained ......."
"remaining ......"
"remains ........"
"remember ......."
"remote ........."
"remove ........."
"removed ........"
"repeat ........."
"repeated ......."
"replace ........"
"reply .........."
"report ........."
"reported ......."
"reports ........"
"represents ....."
"request ........"
"requested ......"
"require ........"
"required ......."
"requires ......."
"requiring ......"
"reserved ......."
"resource ......."
"resources ......"
"respect ........"
"response ......."
"rest ..........."
"result ........."
"resultant ......"
"resulted ......."
"resulting ......"
"results ........"
"return ........."
"returned ......."
"rev ............"
"revealed ......."
"review ........."
"reviews ........"
"revised ........"
"right .........."
"rights ........."
"ring ..........."
"rise ..........."
"role ..........."
"rom ............"
"root ..........."
"rule ..........."
"rules .........."
"run ............"
"running ........"
"said ..........."
"sake ..........."
"same ..........."
"san ............"
"sat ............"
"saturday ......."
"save ..........."
"saved .........."
"saw ............"
"say ............"
"saying ........."
"says ..........."
"search ........."
"second ........."
"section ........"
"secure ........."
"see ............"
"seeing ........."
"seek ..........."
"seem ..........."
"seemed ........."
"seems .........."
"seen ..........."
"sees ..........."
"sein ..........."
"seine .........."
"select ........."
"selected ......."
"self ..........."
"sell ..........."
"selves ........."
"send ..........."
"sending ........"
"sends .........."
"sense .........."
"sent ..........."
"sep ............"
"separate ......."
"separately ....."
"september ......"
"series ........."
"seriously ......"
"server ........."
"servers ........"
"service ........"
"services ......."
"serving ........"
"ses ............"
"set ............"
"setting ........"
"seven .........."
"seventh ........"
"seventy ........"
"several ........"
"shall .........."
"shalt .........."
"share .........."
"she ............"
"ship ..........."
"short .........."
"shortly ........"
"should ........."
"shouldn ........"
"show ..........."
"showed ........."
"showing ........"
"shown .........."
"shows .........."
"shtml .........."
"sich ..........."
"side ..........."
"sie ............"
"similar ........"
"simple ........."
"simply ........."
"sin ............"
"since .........."
"sind ..........."
"single ........."
"site ..........."
"sites .........."
"six ............"
"sixth .........."
"sixty .........."
"size ..........."
"skip ..........."
"slight ........."
"slightly ......."
"slowly ........."
"small .........."
"smaller ........"
"software ......."
"solely ........."
"some ..........."
"somebody ......."
"somehow ........"
"someone ........"
"something ......"
"sometime ......."
"sometimes ......"
"somewhat ......."
"somewhere ......"
"son ............"
"soon ..........."
"sooner ........."
"sort ..........."
"sought ........."
"sound .........."
"sounds ........."
"source ........."
"sources ........"
"space .........."
"spacer ........."
"speak .........."
"speaks ........."
"special ........"
"specific ......."
"specificall ...."
"specifically ..."
"specify ........"
"specifying ....."
"speech ........."
"speed .........."
"spite .........."
"spoken ........."
"stand .........."
"standard ......."
"standing ......."
"stands ........."
"star ..........."
"start .........."
"started ........"
"starting ......."
"starts ........."
"state .........."
"stated ........."
"statement ......"
"states ........."
"stating ........"
"stay ..........."
"still .........."
"stood .........."
"stop ..........."
"stops .........."
"store .........."
"stories ........"
"story .........."
"strong ........."
"student ........"
"studies ........"
"study .........."
"stuff .........."
"style .........."
"sub ............"
"subject ........"
"subjects ......."
"submit ........."
"subscribe ......"
"such ..........."
"suddenly ......."
"suffer ........."
"suggest ........"
"suggestion ....."
"suggests ......."
"summary ........"
"sun ............"
"sunday ........."
"super .........."
"support ........"
"supported ......"
"supports ......."
"suppose ........"
"supposed ......."
"sur ............"
"sure ..........."
"surely ........."
"system ........."
"systems ........"
"table .........."
"tabs ..........."
"take ..........."
"taken .........."
"takes .........."
"taking ........."
"talk ..........."
"talked ........."
"talks .........."
"teach .........."
"tell ..........."
"telling ........"
"tells .........."
"ten ............"
"tend ..........."
"tended ........."
"tending ........"
"tends .........."
"tenth .........."
"term ..........."
"terms .........."
"text ..........."
"texts .........."
"than ..........."
"thank .........."
"thanks ........."
"that ..........."
"the ............"
"thee ..........."
"their .........."
"theirs ........."
"them ..........."
"theme .........."
"themselves ....."
"then ..........."
"thence ........."
"there .........."
"thereby ........"
"therefore ......"
"therein ........"
"thereof ........"
"these .........."
"they ..........."
"thine .........."
"thing .........."
"things ........."
"think .........."
"thinking ......."
"thinks ........."
"third .........."
"thirty ........."
"this ..........."
"those .........."
"thou ..........."
"though ........."
"thought ........"
"thousand ......."
"three .........."
"through ........"
"throughout ....."
"thu ............"
"thursday ......."
"thus ..........."
"thy ............"
"tiff ..........."
"till ..........."
"time ..........."
"times .........."
"title .........."
"toc ............"
"today .........."
"together ......."
"told ..........."
"too ............"
"took ..........."
"top ............"
"topics ........."
"total .........."
"totally ........"
"toward ........."
"towards ........"
"transfer ......."
"treat .........."
"treats ........."
"tried .........."
"true ..........."
"truly .........."
"truth .........."
"try ............"
"trying ........."
"tue ............"
"tuesay ........."
"tuesday ........"
"turn ..........."
"turned ........."
"turns .........."
"twelfth ........"
"twelve ........."
"twentieth ......"
"twenty ........."
"twice .........."
"two ............"
"txt ............"
"type ..........."
"types .........."
"typical ........"
"uber ..........."
"unable ........."
"und ............"
"under .........."
"understand ....."
"understanding .."
"understands ...."
"understood ....."
"une ............"
"unknown ........"
"unless ........."
"unlimited ......"
"uns ............"
"until .........."
"unto ..........."
"update ........."
"updated ........"
"upon ..........."
"upper .........."
"url ............"
"urls ..........."
"usage .........."
"use ............"
"used ..........."
"useful ........."
"user ..........."
"users .........."
"uses ..........."
"using .........."
"usual .........."
"usually ........"
"utterly ........"
"value .........."
"variety ........"
"various ........"
"version ........"
"versions ......."
"very ..........."
"via ............"
"video .........."
"view ..........."
"viewed ........."
"viewer ........."
"views .........."
"virtually ......"
"vision ........."
"visit .........."
"visitor ........"
"visitors ......."
"visual ........."
"vol ............"
"volume ........."
"von ............"
"wait ..........."
"waiting ........"
"want ..........."
"wanted ........."
"wants .........."
"war ............"
"was ............"
"wasn ..........."
"watch .........."
"wave ..........."
"way ............"
"ways ..........."
"web ............"
"webmaster ......"
"webring ........"
"website ........"
"wed ............"
"wednesday ......"
"week ..........."
"weekly ........."
"weeks .........."
"welcome ........"
"well ..........."
"wenn ..........."
"went ..........."
"were ..........."
"what ..........."
"whatever ......."
"when ..........."
"whenever ......."
"where .........."
"whereas ........"
"whereby ........"
"wherein ........"
"wherever ......."
"whether ........"
"which .........."
"while .........."
"who ............"
"whole .........."
"whom ..........."
"whose .........."
"why ............"
"wide ..........."
"widely ........."
"wie ............"
"will ..........."
"wilt ..........."
"win ............"
"window ........."
"windows ........"
"wir ............"
"wish ..........."
"with ..........."
"within ........."
"without ........"
"won ............"
"word ..........."
"words .........."
"work ..........."
"working ........"
"works .........."
"worse .........."
"worth .........."
"would .........."
"wouldn ........."
"write .........."
"writes ........."
"writing ........"
"writings ......."
"written ........"
"wrote .........."
"www ............"
"yahoo .........."
"year ..........."
"years .........."
"yes ............"
"yet ............"
"you ............"
"your ..........."
"yours .........."
"yourself ......."
"yourselves ....."
"zero ..........."
"zip ............"
"~ .............." //top sentinel
};

	//Argv[1] could be an existing or new filename to hold URL text lines.
	//Lately, I've been refering to the ScriptFilename file as the 'LIST'.
	//Each URL is pure text, like http://locn/path;param?query (no #frag).
	//After each url line may be annotation lines starting with one space.

char * ScriptFilename = 0;

	//Argv[2] must contain flag letter argument (-a, -f, -m, -q, -w).

char CmdLineLetter = ' ';
char ProgramIsDoingFetch = 0; //For the -F flag
char ProgramIsDoingWrite = 0; //For the -W flag
char ProgramIsDoingMkdir = 0; //For the -F and -M flags

char QueryTopicString [260];

	//The following array contains 2 strings for each search engine:
	//1) The URL ready to have a QueryTopicString added at the end,
	//2) a title string to use when building list file for -Q flag.
	//I'll take off the * for the most popular and generic engines.

char * SearchEngineQueriesAndTitles [] = {
	"http://av.yahoo.com/bin/query?hc=0&hs=0&p=",
	"Your query for Yahoo! AltaVista",
	"http://berg.cs.colostate.edu:1969/nph-search?classic=on&t1=x&Boolean=AND&Hits=10&Mode=MakePlan&df=normal&AutoStep=on&KW=",
	"*Your query for SavvySearch",
	"http://hyperarchive.lcs.mit.edu/cgi-bin/NewSearch?key=",
	"*Your query for HyperArchive",
	"http://ms3.dogpile.com/search?fs=web&ss=stop&to=ten&q=",
	"*Your query for Dogpile",
	"http://newtoo.manifest.com/perl/query?t=",
	"*Your query for What's New Too!",
	"http://search.dejanews.com/dnquery.xp?defaultOp=AND&svcclass=dncurrent&maxhits=20&ST=QS&format=terse&DBS=2&QRY=",
	"*Your query for Deja News - Quick Search",
	"http://search.excite.com/search.gw?trace=a&search=",
	"*Your query for Excite",
	"http://tile.net/search?news=yes&all=yes&words=",
	"*Your query for TILE.NET",
	"http://webcrawler.com/cgi-bin/WebQuery?searchText=",
	"Your query for WebCrawler",
	"http://www.altavista.digital.com/cgi-bin/query?search.x=38&search.y=6&pg=q&what=web&kl=XX&q=",
	"Your query for AltaVista: Simple Query",
	"http://www.cs.ruu.nl/cgi-bin/faqwais?query=",
	"*Your query for Utrecht CS News.Answers FAQ Access Methods",
	"http://www.forumone.com/cgi-prosody/srch?DATABASE=f1&HTTP_PATH=%2Fusr%2Flocal%2Fwww%2Fhttpd%2Fprosody&TERM_1=",
	"*Your query for Eforums",
	"http://www.gold.net/cgi-bin/srch2html?type=Any&field=Any&Location=Any&key=",
	"*Your query for Gold.NET",
	"http://www.goto.com/search-results?Keywords=",
	"*Your query for GoTo",
	"http://www.infoseek.com/Titles?col=WW&sv=IS&lk=noframes&nh=10&qt=",
	"*Your query for Infoseek: The Entertainment Channel",
	"http://www.lycos.com/cgi-bin/pursuit?adv=0&pq=&query=",
	"Your query for Lycos",
	"http://www.mckinley.com/search.gw?c=web&look=magellan&search=",
	"*Your query for Magellan",
	"http://www.metacrawler.com/crawler?method=0&format=1&region=0&rpp=20&timeout=15&hpe=10&general=",
	"*Your query for Metacrawler",
	"http://www.nexor.com/public/aliweb/search/cgi-bin/form?searchtype=Whole+Word&types=Any&titlefield=on&descriptionfield=on&keywordfield=on&showdescription=on&showurl=on&domain=&hits=&query=",
	"*Your query for Nexor",
	"http://www.northernlight.com/nlquery.fcg?si=SOURCE%3D1&cb=0&cc=&us=025&qr=",
	"*Your query for Northern Light",
	"http://www.reference.com/cgi-bin/pn/go.py?searchspace=Mailing+List+Archive&ranking=by+Relevance&querytext=",
	"*Your query for Reference.COM Advanced Search",
	"http://www.search.hotbot.com/hResult.html?SM=MC&DV=730&RG=.com&DC=50&DE=2&OPs=MDRTP&_v=2&DU=days&SW=web&MT=",
	"*Your query for HotBot",
};

	//DiscardingAllText is true for -A when re-parsing local files.

char DiscardingAllText = 1;

//I malloc and realloc to grow this array of word item pointers.
//Each item = (one short count + one asciz string + another \0).
//These strings end in \0\0 to speed comparison with \0\1 words.
char** ppSortedVocabulary = 0;
int nMallocVocabulary = 0;
int nSortedVocabulary = 0;

	//This holds an insertion-sorted index listing all known URLS.
	//This one is NOT done as a linked-list, like filenames below.

struct UrlPtrs {
	char *pUrl;
	char *pNote;
};
struct UrlPtrs * pSortedUrlPtrs = 0;
int nMallocUrlPtrs = 0;
int nSortedUrlPtrs = 0;
int LastUrlPtrSlot = -1;

	//During resort of local URLs, I could sure use an eXtra int.
struct UrlPtrsX {
	char *pUrl;
	char *pNote;
	int X;
};
struct UrlPtrsX * pResortedUrlPtrs = 0;
int nResortedUrlPtrs = 0;

	//I maintain linked lists by type-casting a (void**) atop data.
	//Unsorted linked list of filenames, after expanding filespecs.

void** FileListHead = 0;
void** FileListTail = (void**) & FileListHead;

	//Now, a linked list of "@..." lines in the order found in LIST.
	//These lines have "@path word word word" to control file names.

void** NamingListHead = 0;
void** NamingListTail = (void**) & NamingListHead;
int NamingListWorstCaseSize = 0;
char* NamingListMatchTerms = 0;
int NamingListTermsCount = 0;
char** NamingListPtrs = 0;

	//Disk read block, tail buffer, conversion buffer, rescan buffer.

char inbuf [4096 + 1024 + 1024];

	//Variables below this point should be cleared after each file.

	//This is an internal representation of parsed URLS: (base or anchor).
	//I shall arbitrarily limit URL segment lengths, elsewhere URL length.

struct UrlParts {
	char scheme [100];
	char netlocn [300];
	char urlpath [300];
	char parameter [300];
	char query [300];
	char fragment [100];
} BaseUrlParts, LinkUrlParts;

	//InsideComment is the only HTML parse state kept between disk blocks.
	//Other HTML tags cause input data to be pushed back until next block.

char InsideComment = 0;

	//DiscardingTextNow merges the global bool with HTML tag factors.

char DiscardingTextNow = DiscardingAllText;

	//DiscardingDueToTag holds mu_* enum of a tag starting discard.

int DiscardingDueToTag = 0;

	//OutputCol is for wrapping text sentence and paragraph output.

int OutputCol = 0;

	//Certain types of markup tags serve as a clue to separate words.

char TagDelimitsWords = 0;

	//UnfinishedToken is to rejoin words split by <FONT> tags, etc.
	//SavedText is an index in UnfinishedToken, to next empty char.

char UnfinishedToken [200];
int SavedText = 0;

	//TitleBuffer holds some of text between <TITLE> and </TITLE>.

char TitleTextBuffer [200];
int TitleTextBufIndex = 0;
char SavingTitleText = 0;

	//AnchorTextBuffer holds some of text between <A> and </A>.

char AnchorTextBuffer [200];
int AnchorTextBufIndex = 0;
char SavingAnchorText = 0;

	//LocalFilename may be current -A input or -F output filename.

char * LocalFilename;

	//Set this global file pointer to save object being fetched.

FILE * ObjectFile = 0;

	//I gather some statistics about the current file.

int SentenceRecognition = 0;
int CountAcceptableIdeas = 0;
int CountAcceptableLinks = 0;
int CountAcceptableWords = 0;

	//This table rids wierd spaces, international characters.
	//Convert actual < and > into \1 and \2 until next parse.
	//The symbols I could not map easily, I changed to a bar.
	//I have seen 0x92 used as apostraphe, make single quote.
	//I have seen 0x93, 94 as smart quote, make single quote.
	//I have seen 0xb4 used as apostraphe, make single quote.

char Simplification [256] = {
'|', '|', '|', '|',  ' ', '|', '|', '|',  '|', ' ', ' ', '|',  ' ', ' ', '|', '|',  //00H = 0
'|', '|', '|', '|',  '|', '|', '|', '|',  '|', '|', '|', '|',  '|', '|', '|', '|',  //10H = 16
' ', '!', '"', '#',  '$', '%', '&', '\'', '(', ')', '*', '+',  ',', '-', '.', '/',  //20H = 32
'0', '1', '2', '3',  '4', '5', '6', '7',  '8', '9', ':', ';',  '\1', '=', '\2', '?', //30H = 48
'@', 'A', 'B', 'C',  'D', 'E', 'F', 'G',  'H', 'I', 'J', 'K',  'L', 'M', 'N', 'O',  //40H = 64
'P', 'Q', 'R', 'S',  'T', 'U', 'V', 'W',  'X', 'Y', 'Z', '[',  '\\',']', '^', '_',  //50H = 80
'`', 'a', 'b', 'c',  'd', 'e', 'f', 'g',  'h', 'i', 'j', 'k',  'l', 'm', 'n', 'o',  //60H = 96
'p', 'q', 'r', 's',  't', 'u', 'v', 'w',  'x', 'y', 'z', '{',  '|', '}', '~', '|',  //70H = 112 bar is 7c
'|', '|', '|', '|',  '|', '|', '|', '|',  '|', '|', '-', '|',  '|', '|', '|', '|',  //80H = 128
'|', '|', '\'','\'', '\'','|', '|', '|',  '|', '|', '|', '|',  '|', '|', '|', '|',  //90H = 144
'|', '|', '|', '|',  '|', '|', '|', '|',  '|', '|', '|', '<',  '~', '-', '|', '|',  //a0H = 160
'|', '|', '2', '3',  '\'','u', '|', '|',  '|', '1', '|', '>',  '|', '|', '|', '|',  //b0H = 176
'A', 'A', 'A', 'A',  'A', 'A', 'A', 'C',  'E', 'E', 'E', 'E',  'I', 'I', 'I', 'I',  //c0H = 192
'D', 'N', 'O', 'O',  'O', 'O', 'O', '*',  'O', 'U', 'U', 'U',  'U', '|', 'P', 's',  //d0H = 208
'a', 'a', 'a', 'a',  'a', 'a', 'a', 'c',  'e', 'e', 'e', 'e',  'i', 'i', 'i', 'i',  //e0H = 224
'd', 'n', 'o', 'o',  'o', 'o', 'o', '/',  'o', 'u', 'u', 'u',  'u', 'a', 'p', 'y',  //f0H = 240
};

	//The following HTML entity names are converted with malice.
	//Each entry must be 8 chars, have a space after match term,
	//have 1-2 replacement character(s) at (word + 6), word + 7,
	//and be perfectly ASCII sorted to permit the binary search.
	//Symbols that were not easy to map become the vertical bar.
	//The dots prevent my editor conversion of 4 spaces to tabs.

char SortedEntityNames [] = {
"! ....  "  //low sentinel
"AElig AE"
"Aacute A"
"Acirc  A"
"Agrave A"
"Aring  A"
"Atilde A"
"Auml   A"
"Ccedil C"
"ETH .. D"
"Eacute E"
"Ecirc  E"
"Egrave E"
"Euml   E"
"Iacute I"
"Icirc  I"
"Igrave I"
"Iuml   I"
"Ntilde N"
"Oacute O"
"Ocirc  O"
"Ograve O"
"Oslash O"
"Otilde O"
"Ouml   O"
"THORN TH"
"Uacute U"
"Ucirc  U"
"Ugrave U"
"Uuml   U"
"Yacute Y"
"aacute a"
"acirc  a"
"acute  |"
"aelig ae"
"agrave a"
"amp .. &"
"aring  a"
"atilde a"
"auml   a"
"brvbar |"
"ccedil c"
"cedil  |"
"cent   |"
"copy   |"
"curren |"
"deg .. |"
"divide /"
"eacute e"
"ecirc  e"
"egrave e"
"eth .. d"
"euml   e"
"frac12 |"
"frac14 |"
"frac34 |"
"gt ... >"
"iacute i"
"icirc  i"
"iexcl  |"
"igrave i"
"iquest |"
"iuml   i"
"laquo  <"
"lt ... <"
"macr   |"
"micro  u"
"middot |"
"nbsp .  "
"not .. !"
"ntilde n"
"oacute o"
"ocirc  o"
"ograve o"
"ordf   |"
"ordm   |"
"oslash o"
"otilde o"
"ouml   o"
"para   |"
"plusmn |"
"pound  |"
"quot   \""
"raquo  >"
"reg .. |"
"sect   |"
"shy .. |"
"sup1   1"
"sup2   2"
"sup3   3"
"szlig sz"
"thorn th"
"times  *"
"uacute u"
"ucirc  u"
"ugrave u"
"uml .. |"
"uuml   u"
"yacute y"
"yen .. |"
"yuml   y"
"~ ....  " //top sentinel
};

	//The following HTML tag names constitute my recognition universe.
	//They must be perfectly ASCII sorted to permit the binary search.
	//"!doctype", !--, and variants !-, !<space> are not handled here.
	//Do not let SortedTagNames array differ from MarkUps enumeration.

char * SortedTagNames [] = {
"!", //low sentinel
"a",
"abbrev",
"acronym",
"address",
"align",
"applet",
"area",
"arg",
"au",
"author",
"b",
"banner",
"base",
"basefont",
"bdo",
"bgsound",
"big",
"blink",
"blockquote",
"body",
"bq",
"br",
"caption",
"center",
"cite",
"code",
"col",
"colgroup",
"credit",
"dd",
"del",
"dfn",
"dir",
"div",
"dl",
"dt",
"em",
"embed",
"fig",
"fn",
"font",
"form",
"frame",
"frameset",
"h1",
"h2",
"h3",
"h4",
"h5",
"h6",
"head",
"hr",
"html",
"i",
"image",
"img",
"input",
"ins",
"isindex",
"kbd",
"lang",
"lh",
"li",
"link",
"listing",
"map",
"marquee",
"menu",
"meta",
"nextid",
"nobr",
"noembed",
"noframes",
"noscript",
"note",
"object",
"ol",
"op",
"option",
"overlay",
"p",
"param",
"person",
"plaintext",
"pre",
"q",
"quote",
"s",
"samp",
"script",
"select",
"small",
"span",
"strike",
"strong",
"style",
"sub",
"sup",
"tab",
"table",
"tbody",
"tc",
"td",
"text",
"textarea",
"tfoot",
"th",
"thead",
"title",
"tr",
"tt",
"u",
"ul",
"var",
"wbr",
"xmp",
"~", //top sentinel
//no entry needed here for mu_Comment enum
};

	//Do not let SortedTagNames array differ from MarkUps enumeration.
	//Do not let MarkupTagTypes array differ from MarkUps enumeration.

enum MarkUps {
mu_lowsentinel = 0,
mu_a,
mu_abbrev,
mu_acronym,
mu_address,
mu_align, //common invalid tag
mu_applet,
mu_area,
mu_arg,
mu_au,
mu_author, //common invalid tag
mu_b,
mu_banner,
mu_base,
mu_basefont,
mu_bdo,
mu_bgsound,
mu_big,
mu_blink,
mu_blockquote,
mu_body,
mu_bq,
mu_br,
mu_caption,
mu_center,
mu_cite,
mu_code,
mu_col,
mu_colgroup,
mu_credit,
mu_dd,
mu_del,
mu_dfn,
mu_dir,
mu_div,
mu_dl,
mu_dt,
mu_em,
mu_embed,
mu_fig,
mu_fn,
mu_font,
mu_form,
mu_frame,
mu_frameset,
mu_h1,
mu_h2,
mu_h3,
mu_h4,
mu_h5,
mu_h6,
mu_head,
mu_hr,
mu_html,
mu_i,
mu_image, //common invalid tag
mu_img,
mu_input,
mu_ins,
mu_isindex,
mu_kbd,
mu_lang,
mu_lh,
mu_li,
mu_link,
mu_listing,
mu_map,
mu_marquee,
mu_menu,
mu_meta,
mu_nextid,
mu_nobr,
mu_noembed,
mu_noframes,
mu_noscript,
mu_note,
mu_object,
mu_ol,
mu_op,
mu_option,
mu_overlay,
mu_p,
mu_param,
mu_person,
mu_plaintext,
mu_pre,
mu_q,
mu_quote, //common invalid tag
mu_s,
mu_samp,
mu_script,
mu_select,
mu_small,
mu_span,
mu_strike,
mu_strong,
mu_style,
mu_sub,
mu_sup,
mu_tab,
mu_table,
mu_tbody,
mu_tc,
mu_td,
mu_text, //common invalid tag
mu_textarea,
mu_tfoot,
mu_th,
mu_thead,
mu_title,
mu_tr,
mu_tt,
mu_u,
mu_ul,
mu_var,
mu_wbr,
mu_xmp,
mu_topsentinel,
mu_Comment //This one is also used to get past !doctype...
};

	//Do not let MarkupTagTypes array differ from MarkUps enumeration.
	//Bit meanings used in MarkupTagTypes:
	// 1 = This tag separates words, in case no other whitespace.
	// 2 = This tag separates concepts (sentences or paragraphs).
	// 4 = This tag may require URL processing. (A, AREA, BASE, FRAME)
	// 8 = TAG /TAG delimit content to discard. (APPLET, OBJECT, SCRIPT, STYLE)
	//16 = This tag may imply other processing. (IMG or IMAGE, TITLE, /TITLE)

char MarkupTagTypes [] = {
-1, //mu_lowsentinel = 0,
1|4, //mu_a, //Extract <A HREF=URL>, keep URL indications in the text.
1, //mu_abbrev,
1, //mu_acronym,
1|2, //mu_address,
1|2, //mu_align,
1|2|8, //mu_applet, //Discard all content from <APPLET> to </APPLET>.
1|2|4, //mu_area, //Extract <AREA HREF=URL>, keep URL indications in the text.
1, //mu_arg,
1, //mu_au,
1, //mu_author,
1, //mu_b,
1|2, //mu_banner,
1|2|4, //mu_base, //Extract <BASE HREF=URL>, keep URL to resolve local URLs.
0, //mu_basefont,
0, //mu_bdo,
0, //mu_bgsound,
0, //mu_big,
0, //mu_blink,
1|2, //mu_blockquote,
1|2, //mu_body,
1|2, //mu_bq,
1|2, //mu_br,
1|2, //mu_caption,
1|2, //mu_center,
1, //mu_cite, //might insert a single quote, if not quoted
1, //mu_code,
1|2, //mu_col,
1|2, //mu_colgroup,
0, //mu_credit,
1, //mu_dd,
0, //mu_del,
1, //mu_dfn,
1|2, //mu_dir,
1|2, //mu_div,
1|2, //mu_dl,
1|2, //mu_dt,
1, //mu_em,
0, //mu_embed,
0, //mu_fig,
0, //mu_fn,
0, //mu_font,
1|2, //mu_form, //Might keep some sort of FORM indications in the text.
1|2|4, //mu_frame, //Extract <FRAME SRC=URL>, keep URL indications in the text.
1|2, //mu_frameset,
1|2, //mu_h1,
1|2, //mu_h2,
1|2, //mu_h3,
1|2, //mu_h4,
1|2, //mu_h5,
1|2, //mu_h6,
1|2|8, //mu_head, //Discard all content from <HEAD> to </HEAD>.
1|2, //mu_hr,
1|2, //mu_html,
1, //mu_i,
1|2|16, //mu_image, //(Misspelled) Extract <IMAGE ALT=TEXT> for the text.
1|2|16, //mu_img, //Extract <IMG ALT=TEXT> for the text.
1|2, //mu_input,
0, //mu_ins,
1|2, //mu_isindex, //Might extract <ISINDEX ACTION="URL?...">
0, //mu_kbd,
0, //mu_lang,
1|2, //mu_lh,
1|2, //mu_li,
1|2, //mu_link, //Might extract <LINK REL="INDEX" HREF="...">
1|2, //mu_listing,
1|2, //mu_map,
1|2, //mu_marquee,
1|2, //mu_menu,
0, //mu_meta,
0, //mu_nextid,
1|2, //mu_nobr,
1|2, //mu_noembed,
1|2, //mu_noframes,
1|2, //mu_noscript,
1|2, //mu_note,
1|2|8, //mu_object, //Discard all content from <OBJECT> to </OBJECT>.
1|2, //mu_ol,
0, //mu_op,
1|2, //mu_option, //Might extract <OPTION VALUE="(possible URL)">
0, //mu_overlay,
1|2, //mu_p,
1|2, //mu_param, //Might extract <PARAM VALUE="(possible URL)">
0, //mu_person,
1|2, //mu_plaintext,
1|2, //mu_pre,
1, //mu_q,
1|2, //mu_quote,
0, //mu_s,
1|2, //mu_samp,
1|2|8, //mu_script, //Discard all content from <SCRIPT> to </SCRIPT>.
1|2, //mu_select,
0, //mu_small,
0, //mu_span,
0, //mu_strike,
1|2, //mu_strong,
1|2|8, //mu_style, //Discard all content from <STYLE> to </STYLE>.
1, //mu_sub,
1, //mu_sup,
0, //mu_tab,
1|2, //mu_table,
1|2, //mu_tbody,
1|2, //mu_tc,
1|2, //mu_td,
1|2, //mu_text,
1|2, //mu_textarea,
1|2, //mu_tfoot,
1|2, //mu_th,
1|2, //mu_thead,
1|2|16, //mu_title, //Save URL annotation text from <TITLE> to </TITLE>.
1|2, //mu_tr,
1, //mu_tt,
0, //mu_u,
1|2, //mu_ul,
1, //mu_var,
1, //mu_wbr,
1|2, //mu_xmp,
-1, //mu_topsentinel,
1|2, //mu_Comment //This mu_Comment is also used to ignore <!DOCTYPE>
};


void RecursivePathExpansion(int depth, char* GivenPath)
{
	//This adds filenames to the linked list of filenames.
	char* scan = GivenPath;
	char* bgnfile;
	int count = 0;
	struct _finddata_t FindFileData;
	long SearchHandle;
	if(scan [1] == ':' && isalpha(scan [0]))
		scan += 2;
	if(scan [0] == '\\' || scan [0] == '/')
		scan++;
	bgnfile = scan;
	//The idiom 'if(x) do ... while (x);' usually saves a jump opcode
	if(*scan != '\0')
	do
	{
		if(scan [0] == '\\' || scan [0] == '/')
		{
			if(++count > depth)
			{
				char* leftpart = (char*)malloc(scan + 1 - GivenPath);
				if(leftpart == (char*)0)
				{
					fprintf(stderr, "SURF: Malloc failure expanding filenames.\n");
					exit(1);
				}
				strncpy(leftpart, GivenPath, scan - GivenPath);
				leftpart [scan - GivenPath] = '\0';
				if( (SearchHandle = _findfirst( leftpart, &FindFileData )) != -1L )
				{
					do
					{
						if(FindFileData.attrib & _A_SUBDIR)
						{
							if(FindFileData.name [0] != '.'
							||(FindFileData.name [1] != '.'|| FindFileData.name [2] != '\0')
							&& FindFileData.name [1] != '\0')
							{
								int n = (bgnfile - GivenPath) + strlen(FindFileData.name) + strlen(scan) + 1;
								char* thing = (char*)malloc(n);
								if(thing == (char*)0)
								{
									fprintf(stderr, "SURF: Malloc failure expanding filenames.\n");
									exit(1);
								}
								if(GivenPath < bgnfile)
									strncpy(thing, GivenPath, bgnfile-GivenPath);
								strcpy(thing + (bgnfile-GivenPath), FindFileData.name);
								strcat(thing + (bgnfile-GivenPath), scan);
								//printf("I got here for directory '%s'.\n", thing);
								RecursivePathExpansion(depth + 1, thing);
								free(thing);
							}
						}
					} //do...
					while( _findnext( SearchHandle, &FindFileData ) == 0 );
				}
				_findclose( SearchHandle );
				free(leftpart);
				return;
			}
			bgnfile = scan + 1;
		}
		scan++;
	} //do...
	while(*scan != '\0');
	//It's final depth time.
	if( (SearchHandle = _findfirst( GivenPath, &FindFileData )) != -1L )
	{
		do
		{
			if(FindFileData.attrib & _A_SUBDIR)
			{
				if(FindFileData.name [0] != '.'
				||(FindFileData.name [1] != '.'|| FindFileData.name [2] != '\0')
				&& FindFileData.name [1] != '\0')
				{
					int n = (bgnfile - GivenPath) + strlen(FindFileData.name) + 2 + 1;
					char* thing = (char*)malloc(n);
					if(thing == (char*)0)
					{
						fprintf(stderr, "SURF: Malloc failure expanding filenames.\n");
						exit(1);
					}
					if(GivenPath < bgnfile)
						strncpy(thing, GivenPath, bgnfile-GivenPath);
					strcpy(thing + (bgnfile-GivenPath), FindFileData.name);
					strcat(thing + (bgnfile-GivenPath), "\\*");
					//printf("I got here for directory '%s'.\n", thing);
					RecursivePathExpansion(depth + 1, thing);
					free(thing);
				}
			}
			else
			{
				int n = sizeof(void**) + (bgnfile - GivenPath) + strlen(FindFileData.name) + 1;
				char* item = (char*)malloc(n);
				char* fullpath = item + sizeof(void**);
				if(item == (char*)0)
				{
					fprintf(stderr, "SURF: Malloc failure expanding filenames.\n");
					exit(1);
				}
				if(GivenPath < bgnfile)
					strncpy(fullpath, GivenPath, bgnfile-GivenPath);
				strcpy(fullpath + (bgnfile-GivenPath), FindFileData.name);
				//printf("I got here for file '%s'.\n", fullpath);
				*FileListTail = (void**)item;
				FileListTail = (void**)item;
				*FileListTail = (void**)0;
			}
		} //do...
		while( _findnext( SearchHandle, &FindFileData ) == 0 );
	}
	_findclose( SearchHandle );
	return;
}

int IgnoreThisCommonWord(char * word)
{
	//Caller passes a pointer to "alphanumerics\0".
	//If word is in CommonWordList return 1, else 0.
	//This was cloned from PotentialEntityLookup,
	//however, CommonWordList items are 15 chars + 1 space,
	//instead of SortedEntityNames items' 7 chars + 1 space.
	int low = 0;
	int top = (sizeof(CommonWordList) - 1 - 16) >> 4;
	for(;;)
	{
		int mid = low + top >> 1;
		char * eptr = CommonWordList + (mid << 4);
		char * scan = word;
		if(*eptr == *scan)
		{
			//a variant string comparison
			do
			{
				eptr++;
				scan++;
			} //do...
			while(*eptr == *scan);
			//Test for this exact match condition.
			//This time the table entry ends with a space,
			//and the candidate string ends with a null.
			if(*eptr == ' ' && *scan == '\0')
				return 1; //Yes, it is a common word
		}
		if(*eptr > *scan) //Warning - gt/lt on signed chars - ok thru 127.
		{
			//Word was lower. Move down in table.
			if(top == mid)
				return 0; //No, it is an unusual word
			top = mid;
		}
		else
		{
			//Word was higher. Move up in table.
			if(low == mid)
				return 0; //No, it is an unusual word
			low = mid;
		}
	} //for...
}

void InsertWordBeforeWordNumber(int Before, char *pItem)
{
	//Each item = (one short count + one asciz string + another \0).
	//The sorted array of pointers to the items can grow.
	if(nSortedVocabulary == nMallocVocabulary)
	{
		if(nMallocVocabulary == 0)
		{
			//First time, do a malloc.
			nMallocVocabulary = 2048;
			ppSortedVocabulary = (char**)malloc(sizeof(char*) * nMallocVocabulary);
		}
		else
		{
			//Afterwards, do a realloc.
			nMallocVocabulary += 2048;
			ppSortedVocabulary = (char**)realloc(ppSortedVocabulary, sizeof(char*) * nMallocVocabulary);
		}
	}
	if(ppSortedVocabulary == (char**) 0)
	{
		fprintf(stderr, "SURF: Malloc failure creating vocabulary.\n");
		exit(1);
	}
	//To do insertion sort, slide up the end of array to make room.
	//However, if incoming words are sorted there is nothing to do.
	if(Before < nSortedVocabulary)
	{
		char **pPast = ppSortedVocabulary + nSortedVocabulary;
		int n = nSortedVocabulary - Before;
		do {
			*pPast = pPast [-1];
			pPast--;
		} //do...
		while(--n > 0);
	}
	ppSortedVocabulary [Before] = pItem;
	nSortedVocabulary++;
	//Nowadays, I'll use vocabulary size, to not emphasize large files.
	CountAcceptableWords++;
	return;
}

void AddWordToVocabulary(char * start, char * past)
{
	//start contains 1 [a-z] + 0 to 39 [a-z0-9] + (\0 or space at past).
	//I do not require a \0 after, neither will I change that character.
	int n = past - start;
	//fprintf(stderr, "\n(%s)\n", start);
	//The text lines are stored with two null bytes to speed comparison.
	//This candidate text ends with one null, then stops on a 0x01 byte.
	//Toss in a short integer counter atop the item
	char* item = (char*) malloc(sizeof(short) + n + 2);
	if(item == (char*)0)
	{
		fprintf(stderr, "SURF: Malloc failure listing vocabulary.\n");
		exit(1);
	}
	*(short*)item = 1; //Initialize word counter in case is new word.
	memcpy(item + sizeof(short), start, n); //Always n > 0
	item [sizeof(short)+n] = '\0';
	item [sizeof(short)+n+1] = '\1';
	//fprintf(stderr, "AddText: %s.\n", item+sizeof(short));
	{
		//Strictly in the case of equal keys,
		//find any index to a saved item having a word which
		//matches exactly, or exceeds, word copied into Item.
		int low = 0;
		int top = nSortedVocabulary;
		if(top == 0)
		{
			//Repair sentinel byte 2 before storing this item.
			item [sizeof(short)+n+1] = '\0';
			InsertWordBeforeWordNumber(0, item);
			return;
		}
		//Here is the binary search.
		for(;;)
		{
			int mid = low + top >> 1;
			unsigned char * tptr = (unsigned char *)(ppSortedVocabulary [mid] + sizeof(short));
			unsigned char * scan = (unsigned char *)(item + sizeof(short));
			if(*tptr == *scan)
			{
				do
				{
					tptr++;
					scan++;
				} //do...
				while(*tptr == *scan);
			}
			//Test for this exactly matching word condition:
			//In SURF, the pre-sorted word ends with \0, \0,
			//whereas the candidate string ends with \0, \1.
			//The double sentinel saves steps in loop above.
			if(*tptr == '\0' && *scan == '\1')
			{
				//Because ppSortedVocabulary [mid] is an exact match.
				//Count repeated word. Prevent overflow.
				if(++(*(short *)(ppSortedVocabulary [mid])) == 32000)
					--(*(short *)(ppSortedVocabulary [mid]));
				free(item); //Discard malloc'ed copy of repeated word.
				return;
			}
			if(*tptr > *scan)
			{
				//Candidate was lower. Move down in table.
				if(top == mid)
				{
					//Nowhere further to go lower with top.
					//Repair sentinel byte 2 before storing this item.
					item [sizeof(short)+n+1] = '\0';
					//Because ppSortedVocabulary [mid] exceeds Item.
					InsertWordBeforeWordNumber(mid, item);
					return;
				}
				top = mid;
			}
			else
			{
				//Candidate was higher. Move up in table.
				if(low == mid)
				{
					//Nowhere further to go higher with low.
					//Repair sentinel byte 2 before storing this item.
					item [sizeof(short)+n+1] = '\0';
					//Because pItem exceeds ppSortedVocabulary [mid].
					InsertWordBeforeWordNumber(mid + 1, item);
					return;
				}
				low = mid;
			}
		} //for...
	}
	return;
}

void ProcessToken(char * start)
{
	//By token I mean a text sequence: No whitespace, no tags.
	//Token contains only 1+ Simplified[] characters, then \0.
	int n = strlen(start);
	if(SavingTitleText)
	{
		//This diverts text inside <TITLE> ... </TITLE>.
		if(TitleTextBufIndex + n < sizeof(TitleTextBuffer) - 2)
		{
			if(TitleTextBufIndex != 0)
			{
				TitleTextBuffer [TitleTextBufIndex] = ' ';
				TitleTextBufIndex++;
			}
			strcpy(TitleTextBuffer + TitleTextBufIndex, start);
			TitleTextBufIndex += n;
		}
		//Title text is always discarded, being inside <HEAD>...
		return;
	}
	if(SavingAnchorText)
	{
		//This saves a copy of text inside <A> ... </A>.
		if(AnchorTextBufIndex + n < sizeof(AnchorTextBuffer) - 2)
		{
			if(AnchorTextBufIndex != 0)
			{
				AnchorTextBuffer [AnchorTextBufIndex] = ' ';
				AnchorTextBufIndex++;
			}
			strcpy(AnchorTextBuffer + AnchorTextBufIndex, start);
			AnchorTextBufIndex += n;
		}
	}
	if(! DiscardingTextNow) {
		//Place a newline ahead of these characters if needed.
		//This means EOF must place a newline if OutputCol > 0.
		if(OutputCol + n > 60)
		{
			//Lest tokens over 60 chars cause an extra blank line.
			if(OutputCol)
			{
				fputc('\n', ObjectFile);
				OutputCol = 0;
			}
			//Everywhere else except here, the decision to end a
			//concept (sentence or paragraph) should flush token.
		}
		//I prefer one blank to left of even the first word on line.
		fputc(' ', ObjectFile);
		fputs(start, ObjectFile);
		OutputCol += n + 1;

	}
	
	//Now let's re-examine the clear text tokens to do an analysis
	//of vocabulary used in each file to categorize them by content.
	//Also try to discern and count apparent well-formed sentences.
	if(DiscardingDueToTag == 0) //during non-markup clear text
	{
		char WordLike [43]; //Accept words from 1 to 40 chars
		char * from = start;
		char * into;
		for(;;)
		{
			while(!isalpha(*from) && *from != '\0')
			{
				if(*from == '.')
				{
					if(SentenceRecognition >= 3+3) //At least One + 3 words
						CountAcceptableIdeas ++; //Period finished sentence
					SentenceRecognition = 0; //Therefore nothing in progress
				}
				++from;
			}
			if(*from == '\0')
				break;
			//I am definitely sitting on an alpha
			into = WordLike;
			if(SentenceRecognition == 0) //Nothing in progress atop new word.
			{
				if(isupper(*from))
				{
					SentenceRecognition = 2; //Ok, Initial Uppercase letter
				}
				else
				{
					SentenceRecognition = 1; //Enter terminal garbage state
				}
			}
			*into = *from | ' '; //idiom lowercases, ok for [A-Za-z0-9]
			into++;
			from++;
			while(isalpha(*from))
			{
				//I know, this is an excessively restrictive heuristic:
				if(!islower(*from))
					SentenceRecognition = 1; //Enter terminal garbage state
				*into = *from | ' '; //idiom lowercases, ok for [A-Za-z0-9]
				if(into < WordLike + sizeof(WordLike) - 1)
					into++;
				from++;
			}
			if(into > WordLike + 2
			&& into < WordLike + sizeof(WordLike) - 2)
			{
				//Accept words (containing only alphas) from 3 to 40 chars
				//Nowadays, I'll use vocabulary size, instead of bump here.
				//No... CountAcceptableWords ++;
				*into = '\0';
				if(! IgnoreThisCommonWord(WordLike))
	   				AddWordToVocabulary(WordLike, into);
				if(SentenceRecognition >= 2) //Having First Capitalized Word
					SentenceRecognition++; //Above 3, counts all-lower words
			}
			else
			{
				SentenceRecognition = 1; //Enter terminal garbage state
			}
			if(*from == '\0')
				break;
		}
		//No, I decided to count words above, not bytes here:
		//No: CountAcceptableChars += n + 1; //This non-markup clear text + 1 space
	}
	
	return;
}

void AnnotateMarkupIntoText(char * msg)
{
	//Caller passes me everything to put between <>.
	//There may have been some concept being output.
	if(SavedText > 0)
	{
		UnfinishedToken [SavedText] = '\0';
		ProcessToken(UnfinishedToken);
		SavedText = 0;
	}
	if(OutputCol > 0)
	{
		//If there was some current concept,
		//end it, put a blank line after it.
		fputc('\n', ObjectFile);
		fputc('\n', ObjectFile);
		OutputCol = 0;
	}
	//I will indent these markups with tab.
	//I will supply two newlines after msg.
	fprintf(ObjectFile, "\t<%s>\n\n", msg);
	return;
}

void InsertUrlBeforeUrlPtrSlot(int Before, char *pUrl)
{
	//Called from BinaryLookupInsertionSort if no match.
	//The array of "struct UrlPtrs" grows as required.
	if(nSortedUrlPtrs == nMallocUrlPtrs)
	{
		if(nMallocUrlPtrs == 0)
		{
			//First time, do a malloc.
			nMallocUrlPtrs = 2048;
			pSortedUrlPtrs = (struct UrlPtrs *)malloc(sizeof(struct UrlPtrs) * nMallocUrlPtrs);
		}
		else
		{
			//Afterwards, do a realloc.
			nMallocUrlPtrs += 2048;
			pSortedUrlPtrs = (struct UrlPtrs *)realloc(pSortedUrlPtrs, sizeof(struct UrlPtrs) * nMallocUrlPtrs);
		}
	}
	//To do insertion sort, slide up the end of array to make room.
	//However, if incoming lines are sorted there is nothing to do.
	if(Before < nSortedUrlPtrs)
	{
		struct UrlPtrs * pPast = pSortedUrlPtrs + nSortedUrlPtrs;
		int n = nSortedUrlPtrs - Before;
		do {
			pPast[0].pUrl = pPast[-1].pUrl;
			pPast[0].pNote = pPast[-1].pNote;
			pPast--;
		} //do...
		while(--n > 0);
	}
	{
		int n = strlen(pUrl) + 1;
		char* CopyOfUrl = (char*) malloc(n);
		if(CopyOfUrl == (char*)0)
		{
			fprintf(stderr, "SURF: Malloc failure adding URLs to list.\n");
			exit(1);
		}
		strcpy(CopyOfUrl, pUrl);
		pSortedUrlPtrs[Before].pUrl = CopyOfUrl;
		pSortedUrlPtrs[Before].pNote = (char*)0;
	}
	nSortedUrlPtrs++;
	return;
}

void BinaryLookupInsertionSort(char * pUrl)
{
	//Called from AddLinkUrlPartsToUrlList.
	//BinaryLookupInsertionSort always sets LastUrlPtrSlot.
	int low = 0;
	int top = nSortedUrlPtrs - 1;
	//fprintf(stdout, "SURF: %s\n", pUrl);
	if(nSortedUrlPtrs == 0)
	{
		InsertUrlBeforeUrlPtrSlot(0, pUrl);
		LastUrlPtrSlot = 0;
		return;
	}
	//This is the promised first compare to final item.
	//I cloned the loop guts rather than complicate it.
	{
		int sc = strcmp(pSortedUrlPtrs[top].pUrl, pUrl);
		//Test for this exact match condition:
		if(sc == 0)
		{
			//UrlPtrs entry at [top] is an exact match.
			LastUrlPtrSlot = top;
			return;
		}
		if(sc > 0)
		{
			//Candidate was lower. Must fall into the binary search.
		}
		else
		{
			//pUrl exceeds UrlPtrs entry at [top].
			//but there exists no entry at [top + 1].
			InsertUrlBeforeUrlPtrSlot(top + 1, pUrl);
			LastUrlPtrSlot = top + 1;
			return;
		}
	}
   //Here is the binary search.
	for(;;)
	{
		int mid = low + top >> 1;

		int sc = strcmp(pSortedUrlPtrs[mid].pUrl, pUrl);
		//Test for this exact match condition:
		if(sc == 0)
		{
			//UrlPtrs entry at [mid] is an exact match.
			LastUrlPtrSlot = mid;
			return;
		}
		if(sc > 0)
		{
			//Candidate was lower. Move down in table.
			if(top == mid)
			{
				//UrlPtrs entry at [mid] exceeds pUrl.
				//But falls between [mid - 1] and [mid].
				InsertUrlBeforeUrlPtrSlot(mid, pUrl);
				LastUrlPtrSlot = mid;
				return;
			}
			top = mid;
		}
		else
		{
			//Candidate was higher. Move up in table.
			if(low == mid)
			{
				//pUrl exceeds UrlPtrs entry at [mid].
				//But falls between [mid] and [mid + 1].
				InsertUrlBeforeUrlPtrSlot(mid + 1, pUrl);
				LastUrlPtrSlot = mid + 1;
				return;
			}
			low = mid;
		}
	} //for...
}

void MergeNoteAtLastUrlPtrSlot(char *pNote)
{
	//The LastUrlPtrSlot was set by AddLinkUrlPartsToUrlList.
	//Notes will be kept as a linked list, with a (char*) in
	//the first four bytes, or null if this is the last note.
	//When written to scriptfile, each note is on a new line.
	int n = sizeof(char*) + strlen(pNote) + 1;
	char* Item = (char*) malloc(n);
	//fprintf(stdout, "MERG: %s\n", pNote);
	if(Item == (char*)0)
	{
		fprintf(stderr, "SURF: Malloc failure adding URLs to list.\n");
		exit(1);
	}
	*(char**)Item = (char*)0;
	strcpy(Item + sizeof(char*), pNote);
	//Every note starting with a '#' byte must be inserted first.
	//That way you can correctly concatenate multiple LIST files;
	//Not quite, but at least first '#' note will not become '*'.
	//The list file shows leading spaces, but no spaces are here.
	if(*pNote == '#')
	{
		*(char**)Item = pSortedUrlPtrs[LastUrlPtrSlot].pNote;
		pSortedUrlPtrs[LastUrlPtrSlot].pNote = (char*)Item;
	}
	else
	{
		//This is the provision to add this note to end of linked list.
		char ** dad = & pSortedUrlPtrs[LastUrlPtrSlot].pNote;
		if(*dad != (char*)0)
		do
			dad = (char**)*dad;
		while(*dad != (char*)0);
		*dad = Item;
	}
	return;
}

void AddLinkUrlPartsToUrlList()
{
	//Caller may be adding a BASE or Anchor URL,
	//or a Url text line parsed from ScriptFile.
	//In either case URL is now in LinkUrlParts.
	//Perform no sanity checks here. Caller depends on me to set Slot No.
	char UrlText [sizeof(LinkUrlParts)]; //grossly worst case size needed
	char *from, *into = UrlText;
	from = LinkUrlParts.scheme;
	if((*into=*from)!='\0') do from++, into++; while((*into=*from)!='\0');
	*into++ = ':';
	*into++ = '/';
	*into++ = '/';
	from = LinkUrlParts.netlocn;
	if((*into=*from)!='\0') do from++, into++; while((*into=*from)!='\0');

	//I'm going to try appending a '/' to URLS that are just a domain name.
	//But this 3/15/98 change, 2 places, might interfere with url fetching.
	if(*LinkUrlParts.urlpath == '\0'
	&& *LinkUrlParts.parameter == '\0'
	&& *LinkUrlParts.query == '\0')
	{
		*into++ = '/';
		*into = '\0';
	}
	else
	{
		from = LinkUrlParts.urlpath;
		if((*into=*from)!='\0') do from++, into++; while((*into=*from)!='\0');
		from = LinkUrlParts.parameter;
		if((*into=*from)!='\0') do from++, into++; while((*into=*from)!='\0');
		from = LinkUrlParts.query;
		if((*into=*from)!='\0') do from++, into++; while((*into=*from)!='\0');
	}
	BinaryLookupInsertionSort(UrlText);
	return;
}

int LinkUrlPartsIsAcceptable()
{
	//Apply this only to Anchor Urls, not Base Urls,
	//for which I may already have file to describe.
	char * scan = LinkUrlParts.urlpath;
	char * dot = 0;
	char dot0, dot1, dot2, dot3;

	//Pure guesswork here. Your mileage may vary.

	if(strcmp(LinkUrlParts.scheme, "http") != 0)
		return 0; //Currently, only support http

	if(LinkUrlParts.netlocn[0] == '\0')
		return 0; //URL has no domain part

	//However, urlpath may be entirely empty, having only a domain.
	//So, do not apply this planned test:
	//if(LinkUrlParts.urlpath[0] == '\0')
		//return 0; //URL has no path part

	//Look for the final .XXXX extension on the URL filename:
	while(*scan != '\0')
	{
		if(*scan == '.')
			dot = scan + 1; //so actually, dot points to first XXXX letter
		scan++;
	}
	//I can do a little guesswork about the object type found at this URL.
	//This section is likely to become obsolete quickly as new types appear.
	if(scan - dot < 5)
	{
		dot0 = dot[0]|' ';
		dot1 = dot[1]|' ';
		dot2 = dot[2]|' ';
		dot3 = dot[3]|' ';
		switch(scan - dot) {
		case 1:
			//Rid 1: .z,
			if(dot0=='z')
				return 0;
			break;
		case 2:
			//Rid 2: .au, .gz, .pl,
			if(dot0=='a' && dot1=='u')
				return 0;
			if(dot0=='g' && dot1=='z')
				return 0;
			if(dot0=='p' && dot1=='l')
				return 0;
			break;
		case 3:
			//Rid 3: .aif, .cgi, .dll, .gif, .hqx, .jpg,
			// .map, .mp2, .mpg, .pdf, .vmd, .wav, .zip, 
			if(dot0=='a' && dot1=='i' && dot2=='f')
				return 0;
			if(dot0=='c' && dot1=='g' && dot2=='i')
				return 0;
			if(dot0=='d' && dot1=='l' && dot2=='l')
				return 0;
			if(dot0=='g' && dot1=='i' && dot2=='f')
				return 0;
			if(dot0=='h' && dot1=='q' && dot2=='x')
				return 0;
			if(dot0=='j' && dot1=='p' && dot2=='g')
				return 0;
			if(dot0=='m' && dot1=='a' && dot2=='p')
				return 0;
			if(dot0=='m' && dot1=='p' && dot2=='2')
				return 0;
			if(dot0=='m' && dot1=='p' && dot2=='g')
				return 0;
			if(dot0=='p' && dot1=='d' && dot2=='f')
				return 0;
			if(dot0=='v' && dot1=='m' && dot2=='d')
				return 0;
			if(dot0=='w' && dot1=='a' && dot2=='v')
				return 0;
			if(dot0=='z' && dot1=='i' && dot2=='p')
				return 0;
			break;
		case 4:
			//Rid 4: .jpeg, .java,
			if(dot0=='j' && dot1=='p' && dot2=='e' && dot3=='g')
				return 0;
			if(dot0=='j' && dot1=='a' && dot2=='v' && dot3=='a')
				return 0;
			break;
		}
	}
	return 1;
}

void OutputVocabulary(char * CallersBuffer)
{
	//if(CallersBuffer == (char*)0) notes are appended to the URL list.
	//otherwise, I format the top note of URL / HTML file into buffer.
	//Run through the SortedVocabulary in order,
	//Each item = (one short count + one asciz string + another \0).
	//Input words are limited to 1 [a-z] + 0 to 39 [a-z0-9] + \0.
	//Caller must guarantee nSortedVocabulary > 0; saves me one tab;
	char WorkBuffer [120]; //Don't surpass FrequentWords [120] below.
	char * into = WorkBuffer;
	int AddedLines = 0;
	int n = nSortedVocabulary;
	//Processing each Item's contents: create a new sort per counts.
	//Insertion sort place item in a list, descending by counts.
	char **ppItem = ppSortedVocabulary;
	char **ppResortedVocabulary = (char**)malloc(sizeof(char*) * n);
	int nResort = 0;
	if(ppSortedVocabulary == (char**) 0)
	{
		fprintf(stderr, "SURF: Malloc failure outputing vocabulary.\n");
		exit(1);
	}
	do {
		//outer loop processes elements of original list.
		//inner loop inserts that item into the new list.
		//If no favorable strcmp occurs, add item at end.
		//Insert item as soon as a slot compares smaller.
		int itemqty = *(short*)*ppItem;
		int i = 0;
		while(i < nResort)
		{
			if(*(short*)(ppResortedVocabulary [i]) < itemqty)
			{
				//Slide up [i] and all above to make room.
				int k = nResort;
				while (k > i)
				{
					ppResortedVocabulary [k] = ppResortedVocabulary [k-1];
					k--;
				}
				//Put this outer loop's item at vacated i.
				ppResortedVocabulary [i] = *ppItem;
				goto ReSortTwoBreaks;
			}
			i++;
		}
		//Add this outer loop's item to end of list.
		ppResortedVocabulary [nResort] = *ppItem;

	ReSortTwoBreaks: ;

		nResort++;
		ppItem++;
	} //do...
	while(--n > 0);

	//Processing each resorted Item's contents.
	//List the words to work buffer, stop at about 100+ columns.
	n = 0;
	do {
		char * from = ppResortedVocabulary[n] + sizeof(short);
		if(into > WorkBuffer)
			*into++ = ' ';
		if((*into=*from)!='\0') do from++, into++; while((*into=*from)!='\0');
		if(into > WorkBuffer + 60)  //Loop could go 1 word beyond, ~40 chars.
		{
			if(CallersBuffer == (char*)0)
			{
				if(ProgramIsDoingFetch)
					fprintf(stdout, "%s\r\n", WorkBuffer);
				MergeNoteAtLastUrlPtrSlot(WorkBuffer);
			}
			else if(AddedLines == 0)
			{
				strcpy(CallersBuffer, WorkBuffer);
			}
			into = WorkBuffer;
			//The 2/98 version added 3 lines of words, but I think
			//that made the URL list files too huge and unreadable.
			if(++AddedLines == 2)
				break;
		}
	} //do...
	while(++n < nSortedVocabulary);
	//Attach any short line for files with short vocabulary lists
	if(into > WorkBuffer)
	{
		if(CallersBuffer == (char*)0)
		{
			if(ProgramIsDoingFetch)
				fprintf(stdout, "%s\r\n", WorkBuffer);
			MergeNoteAtLastUrlPtrSlot(WorkBuffer);
		}
		else if(AddedLines == 0)
		{
			strcpy(CallersBuffer, WorkBuffer);
		}
	}

	//I am finished with the resorting vector.
	free(ppResortedVocabulary);
	return;
}

void ClearVocabulary()
{
	//Each item = (one short count + one asciz string + another \0).
	//Run through the SortedVocabulary in order,
	//freeing each item's memory.
	int n = nSortedVocabulary;
	if(n > 0)
	{
		char **ppItem = ppSortedVocabulary;
		do {
			free(*ppItem);
			ppItem++;
		} //do...
		while(--n > 0);
		//Restore the empty condition of the list
		free(ppSortedVocabulary);
		nMallocVocabulary = 0;
		ppSortedVocabulary = (char**)0;
		nSortedVocabulary = 0;
	}
	return;
}

void ProcessForEndOfInputFile()
{
	//There may have been some concept being output.
	//I put @ in eof comment, to stop mailto search.
	if(! DiscardingAllText)
	{
		//AnnotateMarkupIntoText will clean up any final concepts.
		AnnotateMarkupIntoText("!-- ######## ######## @End ######## ######## --");
	}

	//Now is the time to record facts about this HTML file I just processed:
	memcpy( (void*) & LinkUrlParts, (void*) & BaseUrlParts, sizeof(LinkUrlParts));

	//Before calling AddLinkUrlPartsToUrlList, make sure any prior added URL
	//got a note, so there will be an asterisk to inhibit further fetching.
	if(nSortedUrlPtrs > 0
	&& pSortedUrlPtrs[LastUrlPtrSlot].pNote == (char*)0)
	{
        //So *(?) after my BASE URL means URL disagreed with file's BASE tag.
		MergeNoteAtLastUrlPtrSlot("*(?)");
	}

	AddLinkUrlPartsToUrlList(); //This is the BASE Url for LocalFilename.
	{
		//Because I have just parsed the file and have the
		//authoritative word about its contents, if there
		//were other <A> Url notes, free those mallocs now.
		char * Item = pSortedUrlPtrs[LastUrlPtrSlot].pNote;
		while(Item != (char*)0)
		{
			char * Next = *(char**)Item;
			free(Item);
			Item = Next;				
		}
		//Be sure to null this head pointer just freed:
		pSortedUrlPtrs[LastUrlPtrSlot].pNote = (char*)0;
	}
	//Always, the first note due to a BASE tag,
	//which BASE tag means that I possess such file,
	//starts with "*", meaning do not refetch file.
	//No, use "#" here, but * on unfetched new links.
	//Then, some terse statistics, local filename.
	{
		char wk [200];
		char * into = wk;
		char * from = LocalFilename; //Normally pathless, but caution.
		*into++ = '#'; //This # means file was fetched, never refetch.
		//Add 3 quality digits -- base-4 logs of Ideas/Links/Chars.
		{
			char c = '0';
			unsigned int n = (unsigned int)CountAcceptableIdeas;
			while(n > 0 && c < '9')
			{
				n >>= 2;
				c++;
			}
			*into++ = c;
		}
		{
			char c = '0';
			unsigned int n = (unsigned int)CountAcceptableLinks;
			while(n > 0 && c < '9')
			{
				n >>= 2;
				c++;
			}
			*into++ = c;
		}
		{
			char c = '0';
			unsigned int n = (unsigned int)CountAcceptableWords;
			while(n > 0 && c < '9')
			{
				n >>= 2;
				c++;
			}
			*into++ = c;
		}
		*into++ = ' ';
		//Add the localfilename being stored, or read.
		while((*into=*from)!='\0' && into < wk + sizeof(wk) - 1)
			from++, into++;
		*into = '\0';
		//The LastUrlPtrSlot was set by AddLinkUrlPartsToUrlList:
		MergeNoteAtLastUrlPtrSlot( wk); //First BASE Url note
	}
	if(TitleTextBufIndex > 0)
	{
		//A second note may be document's cleaned-up <TITLE> info.
		TitleTextBuffer[TitleTextBufIndex] = '\0';
		//The LastUrlPtrSlot was set by AddLinkUrlPartsToUrlList:
		MergeNoteAtLastUrlPtrSlot( TitleTextBuffer); //Second BASE Url note
	}
	else
	{
		MergeNoteAtLastUrlPtrSlot( "No title"); //Second BASE Url note
	}

	if(nSortedVocabulary > 0) //I must guarantee nSortedVocabulary > 0
	{
		//Add 2 lines of the most-used words, each about 60 chars,
		//to annotate the last base href url describing this page.
		OutputVocabulary((char*)0); //to URL list due to NULL pointer
		//Rid the per-input-file word list
		ClearVocabulary();
	}

	//Use EOF opportunity to reset all global HTML parsing state variables.

	InsideComment = 0;
	DiscardingDueToTag = 0;
	DiscardingTextNow = DiscardingAllText;
	OutputCol = 0;
	TagDelimitsWords = 0;
	SavedText = 0;
	TitleTextBufIndex = 0;
	SavingTitleText = 0;
	AnchorTextBufIndex = 0;
	SavingAnchorText = 0;
	LocalFilename = (char*)0;
	SentenceRecognition = 0;
	CountAcceptableIdeas = 0;
	CountAcceptableLinks = 0;
	CountAcceptableWords = 0;


	//Clear the base url before any next file.
	memset( (void*) & BaseUrlParts, 0, sizeof(BaseUrlParts));
	return;
}

void ProcessAsText(char * start, char * beyond)
{
	//I have five callers, all who recognize the end of a run of text.
	//All have space or nonspace at start, all put beyond above start.
	//Callers No. 1 and 5 have \0 byte at beyond.
	//Callers No. 2, 3, 4 have < byte at beyond.
	//I will hurt \0 or <, but then fix it to <.
	char * scan = start;
	*beyond = ' '; //install specific sentinel
	if(SavedText > 0)
	{
		//Either be done with, or add on to a partial token.
		if(*scan == ' ') //This is never the final sentinel space.
		{
			//The saved prior partial token must be ended.
			UnfinishedToken [SavedText] = '\0';
			ProcessToken(UnfinishedToken);
			SavedText = 0;
		}
		else
		{
			if(TagDelimitsWords)
			{
				//Prior partial token was ended by HTML tag.
				//Join up any phrases except alpha to alpha.
				//That is, a delimiting tag ends alpha runs.
				if((!isalpha(*scan) || !isalpha(UnfinishedToken [SavedText-1]))
				&& SavedText < sizeof(UnfinishedToken) - 1)
				do
					UnfinishedToken [SavedText++] = *scan++;
				while(*scan != ' ' && SavedText < sizeof(UnfinishedToken) - 1);
			}
			else
			{
				//Prior partial token gets continued.
				//Append additional characters to it.
				if(SavedText < sizeof(UnfinishedToken) - 1) //Need not ask if *scan != ' '.
				do
				{
					UnfinishedToken [SavedText++] = *scan++;
				} //do...
				while(*scan != ' ' && SavedText < sizeof(UnfinishedToken) - 1);
			}
			//Now suppose this phrase is also a final token as below.
			//In that case, don't yet output unfinished token buffer.
			if(scan != beyond)
			{
				UnfinishedToken [SavedText] = '\0';
				ProcessToken(UnfinishedToken);
				SavedText = 0;
			}
		}
	}
	//Divide the text at all ' ', processing each token.
	//In order to safeguard against multiple whitespace,
	//I will defer applying this skip over spaces until
	//the bottom of the for loop. So, it is needed here.
	if(*scan == ' ' && scan != beyond)
	do
	{
		scan ++;
	} //do...
	while(*scan == ' ' && scan != beyond);
	if(scan != beyond)
	{
		for(;;)
		{
			char * starttoken = scan;
			do
			{
				scan ++;
			} //do...
			while(*scan != ' ');
			*scan = '\0'; //Null terminate the token
			// Note that SavedText = 0 when entering next if stmt.
			if(scan == beyond
			&& scan - starttoken <= sizeof(UnfinishedToken) - 1)
			{
				//Save final word as a possibly partial token.
				//That means EOF must trigger the flush of it.
				//Append to partial token, unless was too big.
				scan = starttoken; //Need not ask anything before this do.
				do
					UnfinishedToken [SavedText++] = *scan++;
				while(*scan != '\0'); //Need not test SavedText overflow.
				//Notice that I have not terminated UnfinishedToken yet.
				TagDelimitsWords = 0;
				break; //because scan had equaled beyond
			}
			else
			{
				//PrAsTxt is Not a fit final token.
				ProcessToken(starttoken);
				//However, it could have been a final token, so be careful.
				if(scan != beyond) //Need not ask if *scan == ' '
				do
				{
					scan ++;
				} //do...
				while(*scan == ' ' && scan != beyond);
				if(scan == beyond)
					break;
			}
		}
	}
	*beyond = '<'; //fix this in case partial tag gets pushed back.
	return;
}

int GetLinkUrlParts(char * url, char * beyond)
{
	//Url is non-empty \0 terminated param value
	//of BASE HREF, A HREF, AREA HREF, FRAME SRC.
	//Both callers are ok that I modify sentinel.
	char *start = url;
	char *cease = beyond;
	char overflowed = 0;

	//I fill out data of the LinkUrlParts global,
	//perhaps using data taken from BaseUrlParts.

	memset( (void*) & LinkUrlParts, 0, sizeof(LinkUrlParts));

	//A parse recommendation is given in RFC1808.

	// 2.4.1 Everything to the right of leftmost '#' is a <fragment>.
	// I don't care about <fragment>, but I will extract them anyway.
	// In same loop I will scan and re-pack out any space characters.
	{ //I need not ask if(start != cease)
		char * from = start;
		char * into;
		*cease = '#'; //install my loop sentinel
		//Enter a first phase that scans without copying as into == from.
		if(*from != '#' && *from != ' ')
		do {
			from++;
		} //do...
		while (*from != '#' && *from != ' ');
		into = from; //to repair cease, even if do not do next phase.
		//Enter a second phase using a similar loop only if required.
		if(*from == ' ')
		{
			//I do this because HTML documents sometimes get word wrapped.
			//There might remain spurious hyphens to left of these spaces.
			do {
				if(*from != ' ') //Omit space characters before the #
					*into++ = *from;
				from++;
			} //do...
			while (*from != '#');
		}
		if(from != cease)
		{
			//We have found a #<fragment> on this URL.
			int n = cease - from;
			if(n > sizeof(LinkUrlParts.fragment) - 1)
			{
				n = sizeof(LinkUrlParts.fragment) - 1; //truncate if too long
				// I don't care to set overflowed = 1;
			}
			//This loop will discard the initial #,
			//and it leaves a final destination \0.
			do {
				n--;
				LinkUrlParts.fragment [n] = from [n];
			} while (n > 0);
		}
		cease = into; //Remove the #<fragment> from URL. Also repair slippage.
	}

	// 2.4.2 One or more [a-zA-Z0-9+.-] followed by a ':' is a <scheme>.
	if(start != cease
	&& (isalnum(*start) || *start == '+' || *start == '.' || *start == '-'))
	{
		char * from = start;
		int i = 0;
		*cease = '\0'; //install my loop sentinel
		do
		{
			//ASCII [A-Za-z0-9+.-] can all withstand this lowercasing idiom.
			if(i < sizeof(LinkUrlParts.scheme) - 1) //truncate if too long
			{
				LinkUrlParts.scheme [i] = *from | ' ';
				i++;
			}
			from++;
		} //do...
		while (isalnum(*from) || *from == '+' || *from == '.' || *from == '-');
		if(*from == ':')
		{
			//Syntactically, that was a scheme name. Remove <scheme>: from URL.
			start = from + 1; //Need not ask if from==cease
			if(i == sizeof(LinkUrlParts.scheme) - 1)
			{
				overflowed = 1;
			}
		}
		else
		{
			//Syntactically, that was not a scheme name. Erase it again.
			from = start;
			do LinkUrlParts.scheme [--i] = '\0'; while(i > 0);
		}
	}

	// 2.4.3 If remainder starts "//", everything until "/" is a <netloc>.
	if(start != cease
	&& start[0] == '/'
	&& start[1] == '/')
	{
		char * from = start + 2;
		int i = 0;
		*cease = '/'; //install my loop sentinel
		if (*from != '/')
		do
		{
			if(i < sizeof(LinkUrlParts.netlocn) - 1) //truncate if too long
			{
				//Lowercase net locations (i.e., domain names).
				if(isupper(*from))
					LinkUrlParts.netlocn [i] = *from | ' ';
				else
					LinkUrlParts.netlocn [i] = *from;
				i++;
			}
			from++;
		} //do...
		while (*from != '/');
		start = from; //Remove <netloc> from left of url.
		if(i == sizeof(LinkUrlParts.netlocn) - 1)
		{
			overflowed = 1;
		}
	}

	// 2.4.4 Everything to the right of leftmost '?' is a <query>.
	if(start != cease)
	{
		char * from = start;
		*cease = '?'; //install my loop sentinel
		//We only need the phase that scans
		if(*from != '?')
		do {
			from++;
		} //do...
		while (*from != '?');
		if(from != cease)
		{
			//We have found a ?<query> on this URL.
			int n = cease - from;
			if(n > sizeof(LinkUrlParts.query) - 1)
			{
				n = sizeof(LinkUrlParts.query) - 1; //truncate if too long
				overflowed = 1;
			}
			//This loop will discard the initial ?,
			//and it leaves a final destination \0.
			do {
				n--;
				LinkUrlParts.query [n] = from [n];
			} while (n > 0);
			cease = from; //Remove the ?<query> from URL.
		}
	}

	// 2.4.5 Everything to the right of leftmost ';' is a <param>.
	if(start != cease)
	{
		char * from = start;
		*cease = ';'; //install my loop sentinel
		//We only need the phase that scans
		if(*from != ';')
		do {
			from++;
		} //do...
		while (*from != ';');
		if(from != cease)
		{
			//We have found a ;<parameter> on this URL.
			int n = cease - from;
			if(n > sizeof(LinkUrlParts.parameter) - 1)
			{
				n = sizeof(LinkUrlParts.parameter) - 1; //truncate if too long
				overflowed = 1;
			}
			//This loop will discard the initial ;,
			//and it leaves a final destination \0.
			do {
				n--;
				LinkUrlParts.parameter [n] = from [n];
			} while (n > 0);
			cease = from; //Remove the ;<parameter> from URL.
		}
	}

	// 2.4.6 Whatever is left, except first '/', is URL path.
	// However, keep any '/' to remember if path is absolute.
	if(start != cease)
	{
		int n = cease - start;
		if(n > sizeof(LinkUrlParts.urlpath) - 1)
		{
			n = sizeof(LinkUrlParts.urlpath) - 1; //truncate if too long
			overflowed = 1;
		}
		//This copy does not discard any chars,
		//and it leaves a final destination \0.
		memcpy( (void*) LinkUrlParts.urlpath, (void*) start, n);
	}

	if(overflowed)
	{
		//My callers will not consult LinkUrlParts if I return -1.
		return -1;
	}

	// 3.3 An important observation of RFC1808 is that
	// the BASE href must come from final redirection.
	// That will influence how I perform GetURL stuff.

	return 0; //zero = success
}

int CombineLinkUrlWithBase()
{
	//Here we continue applying rules from section 4 of RFC1808.
	//Link and Base are UrlParts as prepared in GetLinkUrlParts.

	// Step 1 If BASE=empty, URL = absolute.
	// Step 2a If URL=empty, URL = base URL.

	// For SURF rather, if this link's netlocn
	// and urlpath are both empty, ignore link.
	// This adversely affects rfc1808 examples
	// that should inherit whole base net+path,
	// and along with that parameters or query.
	// <URL:?y> isn't <URL:http://a/b/c/d;p?y>.

	if(LinkUrlParts.netlocn [0] == '\0'
	&& LinkUrlParts.urlpath [0] == '\0')
	{
		//This discards links that have no net location and no path part.
		return -1;
	}

	// Step 2b If URL has scheme, URL = absolute.
	// Otherwise, URL inherits scheme from BASE.
	if(LinkUrlParts.scheme [0] != '\0')
	{
		goto TheCheckingAfterTheMerging;
	}
	else
	{
		strcpy(LinkUrlParts.scheme, BaseUrlParts.scheme);
	}

	// Step 3 If URL has netloc, we're all done.
	// Otherwise, URL inherits netloc from BASE.
	if(LinkUrlParts.netlocn [0] != '\0')
	{
		goto TheCheckingAfterTheMerging;
	}
	else
	{
		strcpy(LinkUrlParts.netlocn, BaseUrlParts.netlocn);
	}

	// Step 4 If URL urlpath starts /, we're all done.
	if(LinkUrlParts.urlpath [0] == '/')
	{
		goto TheCheckingAfterTheMerging;
	}

	// Step 5 If URL urlpath is empty, inherit urlpath from BASE.
	if(LinkUrlParts.urlpath [0] == '\0')
	{
		strcpy(LinkUrlParts.urlpath, BaseUrlParts.urlpath);
		//In that case, empty parameter or query also inherit from BASE.
		if(LinkUrlParts.parameter [0] == '\0')
		{
			strcpy(LinkUrlParts.parameter, BaseUrlParts.parameter);
		}
		if(LinkUrlParts.query [0] == '\0')
		{
			strcpy(LinkUrlParts.query, BaseUrlParts.query);
		}
		//And were done.
		goto TheCheckingAfterTheMerging;
	}

	// Looks like some transplant surgery will be required.
	// To recap, link's urlpath is not empty, nor absolute.
	{
		char wk [4 + sizeof(BaseUrlParts.urlpath) + sizeof(LinkUrlParts.urlpath)];
		char c, *into = wk, *stop = wk;
		char *from = BaseUrlParts.urlpath;
		while((*into = c = *from) != '\0')
		{
			into++;
			from++;
			if(c == '/')
				stop = into;
		}
		// Step 6. Remove everything past rightmost "/",
		// or everything if the BASE urlpath has no "/".
		into = stop;

		// Oct.30,1997 version created BASE.COMlinkurl with no /.
		// Remedy that by checking for an empty pathpart of base.
		if(into == wk)
			*into++ = '/';

		// Step 6. Append link URL's urlpath.
		from = LinkUrlParts.urlpath;
		while((*into = *from) != '\0')
		{
			into++;
			from++;
		}

		// Step 6a,b. Remove all path segments of "./", or any final "."
		*into = '\0'; //Fix sentinel
		//Funny how much time I wasted trying to make a slick loop.
		//Now create a straightforward implementation as described.
		into = from = wk;
		while((*into = c = *from) != '\0')
		{
			into++;
			from++;
			if(c == '.')
			{
				//I have the dot.
				if(*from == '/'
				|| *from == '\0')
				{
					//I have the ./ or a final dot.
					if(into == wk + 1
					|| into [-2] == '/')
					{
						//And it was an entire path segment.
						into--; //Remove the already copied dot
						if(*from == '/')
							from++; //Remove the optional slash
						//Wow. Sleeping on it 4 days made it seem easy.
					}
				}
			}
		}

		// Step 6c,d. Remove L-to-R any "<nondotdot>/../" and "<final>/.."
		*into = '\0'; //Fix sentinel
		into = from = wk;
		while((*into = c = *from) != '\0')
		{
			into++;
			from++;
			if(c == '.'
			&& *from == '.')
			{
				//I have the dot-dot.
				if(from [1] == '/'
				|| from [1] == '\0')
				{
					//I have the ../ or a final dot dot.
					//Here, into > wk + 1 would protect
					//into [-2] read, but into > wk + 2
					//guarantees a non-empty left path.
					if(into > wk + 2
					&& into [-2] == '/')
					{
						//I have the slash left of the dot-dot.
						//I must still determine a non-dot-dot
						//and non-empty segment is to the left.
						char * scan = into - 3;
						while(scan >= wk && *scan != '/')
							scan --;
						//If no / was found, scan is at wk - 1.
						//Invert if statement to protect reads.
						if( ! (scan == into - 2 - 3
							&& into [-3] == '.'
							&& into [-4] == '.') )
						{
							into = scan + 1; //Leave any / before path.
							from++; //Remove the still uncopied dot.
							if(*from == '/')
								from++; //Remove the optional slash.
						}
					}
				}
			}
		}
		*into = '\0';
		if(into - wk > sizeof(LinkUrlParts.urlpath) - 1)
		{
			//Combined BASE + LINK url path would overflow.
			return -1;
		}
		strcpy(LinkUrlParts.urlpath, wk);
	}

TheCheckingAfterTheMerging: ;

	// At this point, I verifed all the samples
	// given in rfc1808, save those I mentioned.

	//Ignore all URLS except FTP, HTTP, GOPHER scheme: e.g., MAILTO.
	//Also remove any netlocn:port suffix equal to the default port.
	if(LinkUrlParts.scheme[0] == 'h')
	{
		if(strcmp(LinkUrlParts.scheme, "http") != 0)
			return -1;
		{
			char * atend = LinkUrlParts.netlocn;
			atend += strlen(LinkUrlParts.netlocn);
			if(atend[-3] == ':'
			&& atend[-2] == '8'
			&& atend[-1] == '0')
				*(int*)(atend-3) = 0; //remove :80 from http netlocn
		}
	}
	else if(LinkUrlParts.scheme[0] == 'f')
	{
		if(strcmp(LinkUrlParts.scheme, "ftp") != 0)
			return -1;
		{
			char * atend = LinkUrlParts.netlocn;
			atend += strlen(LinkUrlParts.netlocn);
			if(atend[-3] == ':'
			&& atend[-2] == '2'
			&& atend[-1] == '1')
				*(int*)(atend-3) = 0; //remove :21 from ftp netlocn
		}
	}
	else if(LinkUrlParts.scheme[0] == 'g')
	{
		if(strcmp(LinkUrlParts.scheme, "gopher") != 0)
			return -1;
		{
			char * atend = LinkUrlParts.netlocn;
			atend += strlen(LinkUrlParts.netlocn);
			if(atend[-3] == ':'
			&& atend[-2] == '7'
			&& atend[-1] == '0')
				*(int*)(atend-3) = 0; //remove :70 from gopher netlocn
		}
	} else {
		return -1;
	}

	//Ignore all URLS lacking a net location, or lacking a url path.
	//In fact, make sure here that all net locations start with a /.

	//A base urlpath may be entirely empty, having only a domain,
	//but this routine is only processing Anchor Urls, so we ask:
	//If merged URL's netloc is empty, or not an absolute path...
	if(LinkUrlParts.netlocn [0] == '\0'
	|| LinkUrlParts.urlpath [0] != '/')
	{
		return -1;
	}

	//Convert urlpath starting with the common /%7e or /%7E into /~

	if(LinkUrlParts.urlpath [1] == '%'
	&& LinkUrlParts.urlpath [2] == '7'
	&& (LinkUrlParts.urlpath [3] | ' ') == 'e')
	{
		char * into = LinkUrlParts.urlpath + 1; //Start past the /
		char * from = into + 3;
		*into++ = '~';
		while((*into = *from) != '\0') into++, from++;
	}

	//This url parsed ok; is http, ftp, or gopher;
	//its net location and url path are non-empty.
	return 0;
}

int ProcessURLTags(char * start, char * beyond, int MarkUpEnum)
{
	//Only these starting tags come from loop that finds tags:
	//<A ...>
	//<AREA ...>
	//<BASE ...>
	//<FRAME ...>
	static char AorBhrefText [] = {"base href = "};
	//I shall arbitrarily limit URL length to some convenient length.
	char PrettyUrl [500 + sizeof(TitleTextBuffer)];
	//Start is upon <, beyond is upon >.
	char *from = start;
	char *into;
	//I will hurt >, but yet not fix it.
	*beyond = '\0';
	strcpy(PrettyUrl, AorBhrefText);
	into = PrettyUrl + sizeof(AorBhrefText) - 1; //On the \0
	if(MarkUpEnum == mu_frame)
	{
		//search for " SRC", also require some '='.
		for(;;)
		{
			if(*from <= ' ')
			{
				if(*from == '\0')
					break;
				if(*from == ' '
				&& (from[1] | ' ') == 's'
				&& (from[2] | ' ') == 'r'
				&& (from[3] | ' ') == 'c')
				{
					from += 4;
					if(*from == '=')
						break;
					if(*from == ' ' && from[1] == '=')
					{
						from++;
						break;
					}
				}
			}
			from++;
		}
	}
	else
	{
		//search for " HREF", also require some '='.
		for(;;)
		{
			if(*from <= ' ')
			{
				if(*from == '\0')
					break;
				if(*from == ' '
				&& (from[1] | ' ') == 'h'
				&& (from[2] | ' ') == 'r'
				&& (from[3] | ' ') == 'e'
				&& (from[4] | ' ') == 'f')
				{
					from += 5;
					if(*from == '=')
						break;
					if(*from == ' ' && from[1] == '=')
					{
						from++;
						break;
					}
				}
			}
			from++;
		}
	}
	if(*from == '\0')
	{
		//Need I fix this? No, nor below... *beyond = '>';
		return -1; //This little piggy has no HREF or SRC =.
	}
	from++;
	if(*from == ' ')		//Never multiple whitespaces, thank me.
		from++;
	{
		char c, term = ' '; //Often ' ' between multiple parameters
		char *stopper = PrettyUrl + sizeof(PrettyUrl) - 1
		  - sizeof("> A frame of: x </a\0") - strlen(TitleTextBuffer);
		if(*from == '"'
		|| *from == '\'')
		{
			term = *from;
			from++;
		}
		*beyond = term; //Change sentinel. Look out below!|?
		do
		{
			//this loop must leave room in PrettyUrl to
			//append (a phrase far below + TitleTextBuffer):
			//"> An area of: TitleTextBufferContents </a\0"
			//"> A frame of: TitleTextBufferContents </a\0"
			*into = c = *from;
			from++;
			into++;
		} while(c != term && into < stopper);
		if(c != term)
		{
			//Need I fix this? No, nor below... *beyond = '>';
			return -1;		//This little piggy was too long.
		}
		into--;			//To rid space ' or " terminator.
	}
	if(into == PrettyUrl + sizeof(AorBhrefText) - 1)
	{
		//Need I fix this? No, nor below... *beyond = '>';
		return -1;		//This little piggy was empty.
	}
	*into = '\0';
	if(MarkUpEnum == mu_base)
	{
		*into = ' ';
		into[1] = '\0';
		if(! DiscardingAllText)
		{
			//Annotate base as <base href= http://domain/path/etc/ >.
			//Annotate first, because GetLinkUrlParts is destructive.
			AnnotateMarkupIntoText(PrettyUrl);
		}
		//Canonicalize and store the base url to resolve locals.
		//Reset these so the ordinary link process will suffice.
		memset( (void*) & BaseUrlParts, 0, sizeof(BaseUrlParts));
		if(GetLinkUrlParts(PrettyUrl + sizeof(AorBhrefText) - 1, into) == 0)
		{
			memcpy( (void*) & BaseUrlParts, (void*) & LinkUrlParts, sizeof(BaseUrlParts));
		}
		//Need I fix this? No, nor below... *beyond = '>';
		return -1; //This was a good URL, but the BASE caller doesn't care
	}
	if(! DiscardingAllText)
	{
		//I have changed my scheme, to leave the </a> tag
		//in the midst of the text, but for non-<A> URLs,
		//I must provide a </a> terminator now, lest none.
		if(MarkUpEnum == mu_a)
		{
			into[0] = ' ';
			into[1] = '\0'; //Don't increment into, as into is used below.
		}
		else
		{
			//I think we may be here for only <AREA> or <FRAME>.
			//Annotate other <FRAME> and <AREA> links as either:
			//"> A frame of: ContentsOfTitleTextBuffer </a\0"
			//"> An area of: ContentsOfTitleTextBuffer </a\0"
			//Needing the excess room in PrettyUrl counted far above.
			*into++ = '>';
			*into++ = ' ';
			if(MarkUpEnum == mu_area)
				strcpy(into, "An area of: ");
			else if(MarkUpEnum == mu_frame)
				strcpy(into, "A frame of: ");
			else
				strcpy(into, "A what? of: ");
			into += strlen(into);
   			strcpy(into, TitleTextBuffer);
			into += strlen(into);
			*into++ = ' ';
			*into++ = '<';
			*into++ = '/';
			*into++ = 'a';
			*into = '\0'; //Don't leave home without it.
		}
		PrettyUrl [3] = 'a'; //Change PrettyUrl "BASE HREF" to "...A HREF".
		AnnotateMarkupIntoText(PrettyUrl+3);
	}
	//I annotated first, because GetLinkUrlParts is destructive.
	//Canonicalize and process absolute or (base+relative) URLs.
	if(GetLinkUrlParts(PrettyUrl + sizeof(AorBhrefText) - 1, into) == 0
	&& CombineLinkUrlWithBase() == 0
	&& LinkUrlPartsIsAcceptable()) //This is an A/Area/Frame Anchor URL
	{
		CountAcceptableLinks++;
		//Before calling AddLinkUrlPartsToUrlList, make sure any prior added URL
		//got a note, so there will be an asterisk to inhibit further fetching.
		if(nSortedUrlPtrs > 0
		&& pSortedUrlPtrs[LastUrlPtrSlot].pNote == (char*)0)
		{
			MergeNoteAtLastUrlPtrSlot("*(?)");
		}
		//Fit this fully specified url into the sorted list.
		//This bool would do top only, save ram... if(ListingAllAnchors)
		{
			AddLinkUrlPartsToUrlList(); //This URL from A, Area, Frame, etc.
			//Need I fix this? No, nor above... *beyond = '>';
			return 0; //This was a good URL, so an Anchor caller can divert text
		}
	}
	//Need I fix this? No, nor above... *beyond = '>';
	return -1; //This was not a good, nice, absolute Anchor URL
}

int PotentialTagNameLookup(char * word)
{
	//Caller passes a pointer to a word
	//in buffer, ending in space or \2.
	//May see final \0, need not match.
	//If word is no tag name, return 0.
	//else return index in enumeration.
	int low = 0;
	int top = (sizeof(SortedTagNames)/sizeof(*SortedTagNames) - 1);
	for(;;)
	{
		int mid = low + top >> 1;
		char * tptr = SortedTagNames [mid];
		char * scan = word;
		//My idiom (| ' ') lowercases alphas without hurting numbers.
		if(*tptr == (*scan | ' '))
		{
			//a variant string comparison
			do
			{
				tptr++;
				scan++;
			} //do...
			while(*tptr == (*scan | ' '));
			//Test for this exact match condition.
			//This time the table entry ends with a null,
			//and the candidate string ends with a space,
			//or a \2 standing for the > to be converted.
			if(*tptr == '\0' && (*scan == ' ' || *scan == '\2'))
				return mid;
		}
		if(*tptr > (*scan | ' ')) //Warning - gt/lt on signed chars - ok thru 127.
		{
			//Word was lower. Move down in table.
			if(top == mid)
				return 0;
			top = mid;
		}
		else
		{
			//Word was higher. Move up in table.
			if(low == mid)
				return 0;
			low = mid;
		}
	} //for...
}

char * PotentialEntityLookup(char * word)
{
	//Caller passes a pointer to "alphanumerics\0".
	//If alphanumerics is no entity name, return 0.
	//else return pointer to matching 8 char block.
	int low = 0;
	int top = (sizeof(SortedEntityNames) - 1 - 8) >> 3;
	for(;;)
	{
		int mid = low + top >> 1;
		char * eptr = SortedEntityNames + (mid << 3);
		char * scan = word;
		if(*eptr == *scan)
		{
			//a variant string comparison
			do
			{
				eptr++;
				scan++;
			} //do...
			while(*eptr == *scan);
			//Test for this exact match condition.
			//This time the table entry ends with a space,
			//and the candidate string ends with a null.
			if(*eptr == ' ' && *scan == '\0')
				return SortedEntityNames + (mid << 3);
		}
		if(*eptr > *scan) //Warning - gt/lt on signed chars - ok thru 127.
		{
			//Word was lower. Move down in table.
			if(top == mid)
				return (char*)0;
			top = mid;
		}
		else
		{
			//Word was higher. Move up in table.
			if(low == mid)
				return (char*)0;
			low = mid;
		}
	} //for...
}

int MultiPassInbufConversion(int nbuffered)
{
	//Process input from (inbuf+4)[0] to (inbuf+4)[nbuffered-1].
	//Check for unsafe garbage or binary, returning -1 to abort.
	//Tail of intermediate disk buffer lacking final whitespace,
	//also incomplete tokens, can be copied back down to inbuf+4
	//to join the next input buffer for processing on next loop.
	//Return count of unprocessed characters sitting at inbuf+4.
	int FinalSpace = nbuffered;
	//locate final whitespace (anything <= ' ') to break buffer.
	(inbuf+4)[-1] = ' '; //sentinel
	//Warning - gt/lt on signed chars would NOT be ok: over 127 in next loop:
	do FinalSpace--; while((unsigned char)(inbuf+4)[FinalSpace] > ' ');
	(inbuf+4)[FinalSpace] = ' '; //revise to exact sentinel byte
	//Stop processing unsafe garbage, or probable binary data.
	if(FinalSpace == -1
	|| nbuffered - FinalSpace > 1000)	//because 1Kb unsafe to copy back
		return -1;
	{
		//Scan to simplify character content, check for binary.
		int WierdBytes = 0;
		int i = FinalSpace;
		do
		{
			unsigned char c = (inbuf+4) [i];
			unsigned char d = Simplification [c]; //Warning - array math on UNsigned chars - ok thru 255.
			if(d != c)
			{
				//Replace character with a simplified character.
				(inbuf+4) [i] = d;
				if(d > ' ')
					WierdBytes ++;  //Don't count <, >, tabs, newlines.
			}
		} //do...
		while(--i >= 0);
		//Stop processing unsafe garbage, or probable binary data.
		if((WierdBytes << 2) > FinalSpace)  //over 1/4 of bytes were wierd
			return -1;
	}

	//Process (inbuf+4) [0] through space byte at (inbuf+4) [FinalSpace].
	//During this process convert "&#255;" and "&sup3;" type of entities.
	{
		int PastCount = FinalSpace + 1;
		int i = 0, o = 0, PastAmp;
		char c;
		//This outer loop encloses several processing states.
		for(;;)
		{
			//During the first state, copy until any '&' or any ' '.
			//Trim leading ' ' and duplicate ' ' but keep final ' '.
			if((inbuf [o++] = c = (inbuf+4) [i++]) != '&')
			do
			{
				if(c == ' ')
				{
					if(o == 1
					|| inbuf [o-2] == ' ')
						o--;
					if(i == PastCount) // Watch for all-important ' ' sentinel
						break;
				}
			} //do...
			while((inbuf [o++] = c = (inbuf+4) [i++]) != '&');

			if(i == PastCount)  // Watch for all-important ' ' sentinel
				break;

			//During the next state, process '&' and certain next bytes.
			PastAmp = o;
			//Examine (and keep copying) one next byte.
			inbuf [o++] = c = (inbuf+4) [i++];
			if(isalpha(c))
			{
				//If started off &[a-zA-Z], then accept following [a-zA-Z0-9]*.
				//Keep copying bytes in case not a valid entity name.
				if(isalnum(inbuf [o++] = c = (inbuf+4) [i++]))
				do
				{
				} //do...
				while(isalnum(inbuf [o++] = c = (inbuf+4) [i++]));
				o--; i--;	//put back the non-alnum
				//Now compare to the list of all valid entity names.
				inbuf [o] = '\0';
				{
					//In buffer ends on space, no partial names.
					char * cp = PotentialEntityLookup(inbuf + PastAmp);
					if(cp != (char*)0)
					{
						//Find the desired new chars at (+6,) +7 past ptr.
						o = PastAmp - 1;
						//This cannot grow, as original entity was larger.
						if(cp [6] != ' ')
							inbuf [o++] = cp [6];
						inbuf [o++] = cp [7];
						//Here, I added code to prevent duplicate spaces.
						if(inbuf [o-1] == ' '
						&& (o == 1 || inbuf [o-2] == ' '))
							o--;
						//Consume optional ';' after this valid conversion.
						if(c == ';')
							i++;
					}
				}
			}
			else if(c == '#' && isdigit((inbuf+4) [i]))
			{
				//If started off &#[0-9], then accept following [0-9]*.
				//Keep copying bytes in case not a valid number.
				if(isdigit(inbuf [o++] = c = (inbuf+4) [i++]))
				do
				{
				} //do...
				while(isdigit(inbuf [o++] = c = (inbuf+4) [i++]));
				o--; i--;	//put back the non-digit
				//Force final null for atoi.
				inbuf [o] = '\0';
				{
					//I will accept a number from 1 to 255,
					//but only after I have homogenized it.
					int number = atoi(inbuf + PastAmp + 1);
					if(number >= 1
					&& number <= 255)
					{
						o = PastAmp - 1;
						inbuf [o++] = Simplification [number]; //Warning - number is not a char.
						//Here, I added code to prevent duplicate spaces.
						if(inbuf [o-1] == ' '
						&& (o == 1 || inbuf [o-2] == ' '))
							o--;
						//Consume optional ';' after this valid conversion.
						if(c == ';')
							i++;
					}
				}
			}
			else
			{
				//Else, is no entity, copy verbatim.
				o--; i--;	//put back the non-suitable until next loop
			}
		} //for...
		//Null-terminate the homogenized and converted string in inbuf.
		inbuf [o] = '\0';
		//Henceforth, process data starting at (inbuf+0).
		//This is quantity of bytes to process: return o;
		//That leaves stuff from (inbuf+4) [FinalSpace+1]
		//to (inbuf+4) [nbuffered-1] unused by this loop.
	}
	//Scan buffer to locate and process markup tags versus text runs.
	//If there is a final incomplete tag, push it back for next loop,
	//but not if that area exceeds 1000 bytes or is merely a comment.

	//Use another multi-phase outer loop as done above.
	//Process from (inbuf+0) [0] to null sentinel byte.

	//Recognize these tags: <VALID, </VALID, <!, and any > as an ending.
	{
		char *scan = inbuf;
		char *start = scan;
		int MarkUpEnum = 0;
		if(InsideComment)
		{
			//Resume looking for end of a comment that split buffers.
			InsideComment = 0; //Reset global bool.
			MarkUpEnum = mu_Comment;	//Restore approx state from prior call.
			goto inner2continue;		//I'm just a little sorry.
		}

		//This outer loop encloses several processing states.
		for(;;)
		{
			//start in text mode, passing through all printables.
			while(*scan >= ' ') //Warning - gt/lt on signed chars - ok thru 127.
				scan++;
			//Get rid of the easy \0 end-of-inbuf possibility first.
			if(*scan == '\0')
			{
				//Process the final few words of text in buffer.
				if(scan > start)
					ProcessAsText(start, scan); //This is No. 1 of 5 callers
				break;
			}
			//Nothing matters in text mode except < with a valid tag.
			if(*scan != '\1') //My signal for <
			{
				*scan = '>'; //convert \2 back into >.
				continue;
			}
			*scan = '<'; //convert \1 back into <.
			//Scan is sitting on <. Try to recognize markup.
			if(isalpha(scan [1]))
			{
				//Text not yet scanned may contain [\0\1\2].
				//In buffer ends on space, no partial names.
				MarkUpEnum = PotentialTagNameLookup(scan + 1);
				if(MarkUpEnum == 0)
					continue; //No match to any known [A-Za-z]... tags
				if(scan > start)
				{
					//Process the words of text before this tag.
					ProcessAsText(start, scan); //This is No. 2 of 5 callers
					start = scan;
				}
			}
			else if(scan [1] == '/'
			&& isalpha(scan [2]))
			{
				//Text not yet scanned may contain [\0\1\2].
				//In buffer ends on space, no partial names.
				MarkUpEnum = PotentialTagNameLookup(scan + 2);
				if(MarkUpEnum == 0)
					continue; //No match to any known /[A-Za-z]... tags
				if(scan > start)
				{
					//Process the words of text before this tag.
					ProcessAsText(start, scan); //This is No. 3 of 5 callers
					start = scan;
				}
			}
			else if(scan [1] == '!')
			{
				//It seems unnecessary to be very delicate.
				//Hence, I shall parse any <! as a comment.
				MarkUpEnum = mu_Comment;
				if(scan > start)
				{
					//Process the words of text before this comment.
					ProcessAsText(start, scan); //This is No. 4 of 5 callers
					start = scan;
				}
			}
			else
			{
				//I cannot make a tag of this <. Treat it as text.
				continue;
			}
			//Continue in non-text mode, searching for some ending.
			//Scan==start are sitting on < starting a valid markup.
			scan++;

inner2continue: ;
			while(*scan >= ' ') //Warning - gt/lt on signed chars - ok thru 127.
				scan++;
			//Handle the in-tag \0 end-of-inbuf possibility.
			if(*scan == '\0')
			{
				if(scan - start > 1000)
				{
					//Bypass giant incomplete tag as unsafe data.
					//If it is not a comment, output it to alert.
					if(MarkUpEnum == mu_Comment)
					{
						InsideComment = 1; //simplified global bit
					}
					else
					{
						ProcessAsText(start, scan); //This is No. 5 of 5 callers
					}
				}
				else
				{
					//From *start to *scan-1 inclusive must be
					//copied up, possibly overlapping, to just
					//below the lowest tail byte, which starts
					//currently with (inbuf+4) [FinalSpace+1].
					do (inbuf+4) [FinalSpace--] = *--scan;
					while (start < scan);
				}
				break;
			}
			//Nothing matters in markup mode except >.
			//I did code to honor single and double quotes during markup,
			//but bad quoting causes loss of content. Now I take first >.
			if(*scan != '\2') //My signal for >
			{
				*scan = '<'; //convert \1 back into <.
				goto inner2continue;
			}
			*scan = '>'; //convert \2 back into >.

			//Process this just slightly validated markup tag.
			//Start is sitting on <, and scan is sitting on >.
			char mutt = MarkupTagTypes [MarkUpEnum];
			if(mutt & 1)
			{
				// 1 = This tag separates words, in case no other whitespace.
				TagDelimitsWords = 1;
			}
			if(mutt & 2)
			{
				// 2 = This tag separates concepts (sentences or paragraphs).
				//There may have been some concept being output.
				if(SavedText > 0)
				{
					UnfinishedToken [SavedText] = '\0';
					ProcessToken(UnfinishedToken);
					SavedText = 0;
				}
				if(OutputCol > 0)
				{
					fputc('\n', ObjectFile);
					fputc('\n', ObjectFile);
					OutputCol = 0;
				}
				SentenceRecognition = 0; //Therefore nothing in progress
			}
			if(mutt & 4)
			{
				// 4 = This tag may require URL processing.
				//Interest is (A, AREA, BASE, FRAME).
				if(start [1] != '/')
				{
					//These that do not have a / are starting tags.
					//A return value 0 implies a nice, absolute URL:
					if(ProcessURLTags(start, scan, MarkUpEnum) != -1)
					{
						//That set me up to call MergeNoteAtLastUrlPtrSlot.
						if(MarkUpEnum == mu_a)
						{
							//Start capturing text between <A> ... </A>,
							//which I may record as a note for this URL,
							//if it's first or longer than any old note.
							SavingAnchorText = 1;
							//Put * in first note to keep surf from fetching it.
							AnchorTextBuffer [0] = '*';
							AnchorTextBufIndex = 1;
						}
						else if(pSortedUrlPtrs[LastUrlPtrSlot].pNote == (char*)0)
						{
							//As AREA, BASE, FRAME don't have an </endtag>,
							//I cannot save some text as with <A> ... </A>.
							//If no other notes were recorded for this URL,
							//add "An area of: TitleTextBufferContents"
							//or "A frame of: TitleTextBufferContents".
							char Annotation [40 + sizeof(TitleTextBuffer)];
							//Oops... SURF really goes far without the *'s!
							if(MarkUpEnum == mu_area)
								strcpy(Annotation, "* An area of: ");
							else if(MarkUpEnum == mu_frame)
								strcpy(Annotation, "* A frame of: ");
							else
								strcpy(Annotation, "* A what? of: ");
							strcat(Annotation, TitleTextBuffer);
							MergeNoteAtLastUrlPtrSlot(Annotation);
						}
					}
				}
				else
				{
					//Ending any mutt = 4 tag
					//Syntactically, /area, /frame, or /base could exist,
					//but only the </A> end tag makes sense in HTML text.
					//So, let's take SavingAnchorText as sufficient clue.
					if(SavingAnchorText)
					{
						//There may have been some concept being output.
						//In particular, get the final word before </A>.
						if(SavedText > 0)
						{
							UnfinishedToken [SavedText] = '\0';
							ProcessToken(UnfinishedToken);
							SavedText = 0;
						}
						//End saving of text to annotate an anchor URL.
						SavingAnchorText = 0;
						if(! DiscardingTextNow)
						{
							//Insert " </A>" right in the text,
							//sort of like ProcessToken would.
							if(OutputCol + 5 > 60)
							{
								fputc('\n', ObjectFile);
								OutputCol = 0;
							}
							fputs(" </a>", ObjectFile);
							OutputCol += 5;
						}
						//I first processed UnfinishedToken above.
						//After fussing with this code 3/15/98,
						//I noticed that some old files were missing
						//their </A>, either due to surf, or perhaps
						//due to the HTML author's coding error.
						//Also, empty text would make me skip it here.
						//So, I'll clone this virgin URL annotating
						//and do it before processing a URL, or EOF.
						AnchorTextBuffer[AnchorTextBufIndex] = '\0';
						if(AnchorTextBufIndex > 0)
						{
							//Now the decision to annotate that last URL
							//with the <A>...</A> text shall be that the
							//URL either has no note, or has a note that
							//is shorter than the currently created note.
							//Examine the item holding the first note.
							char* Item = pSortedUrlPtrs[LastUrlPtrSlot].pNote;
							if(Item == (char*)0)
							{
								//Virgin URL.
								MergeNoteAtLastUrlPtrSlot(AnchorTextBuffer);
							}
							else if(*(char**)Item == (char*)0
							&& *(Item + sizeof(char*)) == '*'
							&& strlen(Item + sizeof(char*)) < (unsigned) AnchorTextBufIndex)
							{
								//URL has only one note, and it is shorter.
								//Also, that note like mine has a '*' mark.
								//First discard the older lesser note item.
								free(Item);
								pSortedUrlPtrs[LastUrlPtrSlot].pNote = (char*)0;
								MergeNoteAtLastUrlPtrSlot(AnchorTextBuffer);
							}
						}
					}
				}
			}
			if(mutt & 8)
			{
				// 8 = TAG /TAG delimit content to discard.
				//Interest is (APPLET, OBJECT, SCRIPT, STYLE).
				if(start [1] == '/')
				{
					//End the same type of discard block.
					if(DiscardingDueToTag == MarkUpEnum)
					{
						DiscardingDueToTag = 0;
						DiscardingTextNow = DiscardingAllText;
					}
				}
				else
				{
					//Start an outermost discard block.
					if(DiscardingDueToTag == 0)
					{
						DiscardingDueToTag = MarkUpEnum;
						DiscardingTextNow = 1;
						//There may have been some concept being output.
						//That's easily handled: make sure each 8 has 2.
					}
				}
			}
			if(mutt & 16)
			{
				//16 = This tag may imply other processing.
				//Interest is (TITLE, /TITLE).
				if(MarkUpEnum == mu_title)
				{
					if(start[1] == '/')
					{
						//Ending </Title>.
						if(SavingTitleText)
						{
							SavingTitleText = 0;
							if(! DiscardingAllText
							&& TitleTextBufIndex > 0)
							{
								char wk[sizeof(TitleTextBuffer) + 20];
								TitleTextBuffer[TitleTextBufIndex] = '\0';
								sprintf(wk, "TITLE> %s </TITLE", TitleTextBuffer);
								AnnotateMarkupIntoText( wk);
							}
						}
					}
					else
					{
						//Starting <Title>.
						SavingTitleText = 1;
						TitleTextBufIndex = 0;
					}
				}
			}
			MarkUpEnum = 0; //Having finished with this tag

			//Resume text mode loop, searching for next < startup.
			//Put scan==start just past > ending the valid markup.
			scan++;
			start = scan;
		} //for...
	}
	//That leaves (inbuf+4) [FinalSpace+1] to (inbuf+4) [nbuffered-1] undone.
	(inbuf+4) [nbuffered] = '\0'; //for this printf
	{
		int noffset = nbuffered - 1 - FinalSpace;
		//Non-overlapping copy of leftovers to start of (inbuf+4).
		if(noffset > 0)
			memcpy((inbuf+4), (inbuf+4) + FinalSpace + 1, noffset);
		(inbuf+4) [noffset] = '\0'; //for this printf
		return noffset;
	}
}

void ReadScriptFileToUrlList()
{
	//The ScriptFilename exists. Import lines of it as URLs.
	int hfi = _open(ScriptFilename, _O_BINARY|_O_RDONLY, _S_IREAD|_S_IWRITE);
	if(hfi == -1)
	{
		fprintf(stderr, "SURF: cannot open file %s: %s.\n", ScriptFilename, strerror(errno));
		exit(1);
	}
	//One of my favorite file input methods is to read whole
	//next disk block just past the tail of unprocessed data.
	//It would be even better if I can acheive int alignment.
	int nread;
	int ByteCount = 0;
	int nSkipAFew = 0;
	int nBuffered = 0;
	//--Outer loop--
	//The outer processes disk blocks until end of file.
	for(;;)
	{
		nread = _read(hfi, inbuf + nBuffered, 4096);
		//EOF gives 0; Errors -1.
		if(nread < 0)
		{
			fprintf(stderr, "SURF: error reading file %s: %s.\n", ScriptFilename, strerror(errno));
			exit(1);
		}
		nBuffered += nread;
		ByteCount += nread;
		//Process text from inbuf [nSkipAFew] to inbuf [nBuffered-1]
		if(nBuffered <= nSkipAFew)
		{
			//At EOF with nothing held over
			break;
		}
		{
			char *start = inbuf + nSkipAFew;
			char *stop = inbuf + nBuffered;
			char *scan = start;
			stop[0] = '\r'; //Install loop sentinel 1
			stop[1] = 'Z'; //Install loop sentinel 2
			//--Middle  loop--
			//The middle loop processes this buffer.
			for(;;)
			{
				//--Inner loop--
				//The inner loop looks for a newline.
				if(*scan != '\r' && *scan != '\n')
				{
					do
					{
						scan++;
					} //do...
					while(*scan != '\r' && *scan != '\n');
				}
				else
				{
					//For any empty line, stop adding to last URL.
					//Otherwise, my SURF legends add to prior URL.
					LastUrlPtrSlot = -1;
				}

				//Check for sentinel. If reached end of disk block,
				//this line (start to scan-1) is likely incomplete.
				//If incomplete, some safe size, and !EOF, keep it.
				//I will accept CR alone or LF alone except at the
				//end of the disk block where only CR+LF is enough.

				if(scan >= stop - 1				//Postpone for single CR or LF too.
				&& scan - start < 1204			//Dont postpone if dangerously big.
				&& nread != 0)					//Dont postpone if now end of file.
				{
					//Postpone it's processing until next disk block.
					//Copy the tail end of buffer back for next loop.
					//I hope saved buss cycles warrant the diversion.
					nSkipAFew = start - stop & sizeof(int) - 1;
					nBuffered = nSkipAFew + stop - start;
					if(nBuffered > 0)
					{
						//This might very well be an overlapping copy, however I
						//believe the customary implementation of memcpy is safe.
						memcpy(inbuf, stop - nBuffered, nBuffered);
					}
					//Break out of the middle loop processing this buffer.
					break;
				}
				//This is where to ProcessLineOfText(start, scan);
				if(start < scan)
				{
					char c = *scan; //Save the newline char
					*scan = '\0'; //terminate the entire line
					//fprintf(stdout, "LINE: %s\n", start);
					if(isalnum(*start))
					{
						//Treat this line as a URL
						//This parse has BaseUrlParts full of all zeros.
						if(GetLinkUrlParts(start, scan) == 0
						&& LinkUrlPartsIsAcceptable()) //This is a Scriptfile URL
						{
							//Here, don't worry if the prior URL had no notes
							AddLinkUrlPartsToUrlList();
						}
					}
					else if(start [0] == '@') //Only in first input column
					{
						//Save user's lines of LIST starting with a "@",
						//in order, by appending to tail of linked list.
						//This is a pretty generic linked list block.
						char * TextData = start + 1; //don't store '@'
						int n = sizeof(void**) + strlen(TextData) + 1;
						char* item = (char*)malloc(n);
						char* itemtext = item + sizeof(void**);
						if(item == (char*)0)
						{
							fprintf(stderr, "SURF: Malloc failure saving '@' lines.\n");
							exit(1);
						}
						NamingListWorstCaseSize += n; //loosey goosey
						strcpy(itemtext, TextData);
						//Hook this item to old tail item, or to head if none.
						*NamingListTail = (void**)item;
						//Revise the tail pointer to new tail, i.e., this item.
						NamingListTail = (void**)item;
						//Place a null pointer in new tail, to terminate list.
						*NamingListTail = (void**)0;
					}
					else
					{
						//Strip any type of leading whitespaces.
						while(*start == ' ' || *start == '\t')
							start++;
						//Treat this line as a note continuation line;
						//but only if there were some non-whitespaces.
						if(LastUrlPtrSlot != -1
						&& start < scan)
						{
							//The LastUrlPtrSlot was set by AddLinkUrlPartsToUrlList:
							MergeNoteAtLastUrlPtrSlot(start);
						}
					}
					*scan = c; //Repair the newline char
				}
				//Jump over the CR or LF newline byte
				scan++;
				//Jump by two in order to count CR+LF as one newline.
				if(scan[-1] == '\r' && scan[0] == '\n')
					scan++;
				//check for sentinel
				if(scan >= stop)
				{
					nSkipAFew = 0;
					nBuffered = 0;
					//Break out of the middle loop processing this buffer.
					break;
				}
				start = scan;
			} //for...
		}
		if(nread == 0)
		{
			//At EOF
			break;
		}
	} //for...
	if(_close(hfi) == -1)
	{
		fprintf(stderr, "SURF: error closing file %s: %s.\n", ScriptFilename, strerror(errno));
		exit(1);
	}
	if(nSortedUrlPtrs == 0
	&& ByteCount > 2000)
	{
		//This check is to avoid re-writing non-list files.
		//But, accept a nearly empty list file with 0 urls.
		fprintf(stderr, "SURF: File %s had no 'http://url...' lines.\n", ScriptFilename);
		exit(1);
	}
	return;
}

void SplitOutTheFetchedUrls()
{
	//This was the top half of WriteUrlListToScriptFile(),
	//but I split it off to permit file renaming activity
	//to take place after the link resorting done in here.
	int i, n;
	//Warning: nSortedUrlPtrs could be 0. Therefore I said +1.
	//These two will be cleared in WriteUrlListToScriptFile().
	pResortedUrlPtrs = (struct UrlPtrsX *) malloc (
		sizeof(struct UrlPtrsX) * nSortedUrlPtrs + 1 ); //Worst case
	nResortedUrlPtrs = 0; //To count items relinked.
	if(pResortedUrlPtrs == (struct UrlPtrsX *) 0)
	{
		fprintf(stderr, "SURF: Malloc failure outputing URL list.\n");
		exit(1);
	}
	//Process original list to transfer all the visited ('#') links
	//to a new list, where they are resorted by some other criteria.
	n = 0;
	while(n < nSortedUrlPtrs)
	{
		//This loop examines item: pSortedUrlPtrs[n].
		//Examine the first note for this url, if any.
		//If it's first char is '#', move to new list.
		char * ItemNote1 = pSortedUrlPtrs[n].pNote;
		if(ItemNote1 != (char*)0
		&& *(ItemNote1 + sizeof(char*)) == '#')
		{
			//This insertion sort is ordered by the 3 digits that
			//follow the '#' in first note: (Ideas, Links, Words).
			char *itemkey = ItemNote1 + sizeof(char*) + 1;
			int Quality = itemkey[0] + itemkey[1] + itemkey[2];

			//Inner loop inserts that item into the new list.
			//If no favorable strcmp occurs, add item at end.
			//Insert item as soon as a slot compares smaller.
			int i = 0;
			while(i < nResortedUrlPtrs)
			{
				if(pResortedUrlPtrs[i].X < Quality)
				{
					//Slide up [i] and all above to make room.
					int k = nResortedUrlPtrs;
					while (k > i)
					{
						//Array holds the structs, not pointers to structs.
						//So, copy both members. No, X makes three members.
						pResortedUrlPtrs[k].X = pResortedUrlPtrs[k-1].X;
						pResortedUrlPtrs[k].pUrl = pResortedUrlPtrs[k-1].pUrl;
						pResortedUrlPtrs[k].pNote = pResortedUrlPtrs[k-1].pNote;
						k--;
					}
					//Put this outer loop's item at vacated i.
					pResortedUrlPtrs[i].X = Quality;
					pResortedUrlPtrs[i].pUrl = pSortedUrlPtrs[n].pUrl;
					pResortedUrlPtrs[i].pNote = pSortedUrlPtrs[n].pNote;
					goto ReSortURLTwoBreaks;
				}
				i++;
			}
			//Add this outer loop's item to end of list.
			pResortedUrlPtrs[nResortedUrlPtrs].X = Quality;
			pResortedUrlPtrs[nResortedUrlPtrs].pUrl = pSortedUrlPtrs[n].pUrl;
			pResortedUrlPtrs[nResortedUrlPtrs].pNote = pSortedUrlPtrs[n].pNote;

		ReSortURLTwoBreaks: ;

			nResortedUrlPtrs++;
			//Nullify the moved item out of original list.
			pSortedUrlPtrs[n].pUrl = (char*)0;
			pSortedUrlPtrs[n].pNote = (char*)0;
		}
		n++;
	}

	//Rather than disrupt that work with a bunch of linked lists to
	//keep separate lists of files containing a NamingListPtrs word,
	//Put into .X the lowest index to any equal NamingListPtrs word.

	//Process the new list to re-catalog the most-used words,
	//for a most-used-of-the-most-used-words in LIST and HTM.
	i = 0;
	while(i < nResortedUrlPtrs)
	{
		//Cloned looping from the cloned loop below.
		char * ItemNote = pResortedUrlPtrs[i].pNote;
		int NoteCounter = 0;
		int BestMatch = NamingListTermsCount;
		while(ItemNote != (char*)0)
		{
			char * Next = *(char**)ItemNote;
			//Do not tally first or second notes,
			//which are filename and title lines.
			NoteCounter++;
			if(NoteCounter > 2)
			{
				char * start = ItemNote + sizeof(char*);
				char * past = start;
				//Scan the asciz record to tally each word in it.
				while (*past != '\0')
				{
					past++;
					if(*past <= ' ')
					{
						//There should never be zero length cases,
						//but I see fit to defend my code from it.
						//For example, user may have edited lines.
						if(past > start + 1)
						{
							AddWordToVocabulary(start, past);
							if(BestMatch > 0)
							{
								int j = 0;
								do
								{
									//Oops, this let short words match long:
									//if(strncmp(NamingListPtrs [j+j],
									//start, past-start) == 0)
									char * from = NamingListPtrs [j+j];
									char * into = start;
                                    if(*from == *into)
									{
										do
										{
										    from++; into++;
										} //do...
										while(*from == *into && *from != '\0');
										//Now test comparison loop outcome
										if(*from == '\0' && into == past)
										{
											BestMatch = j;
											break;
										}
									}
								} //do...
								while(++j < BestMatch);
							}
						}
						start = past + 1;
					}
				}
			}
			ItemNote = Next;				
		}
		//Having run through all the notes for this URL,
		//put in .X the index of earliest matching term.
		pResortedUrlPtrs[i].X = BestMatch;
		i++;
	}
	return;
	//Between my return and the WriteUrlListToScriptFile() call,
	//the move (-M) operation might do a bunch of file renaming.
}

void MoveAboutTheFetchedUrls()
{
	//I am located between Split* and Write* because
	//that is the way I am called for the -M command.

	//Process the new list to move any files that are not
	//already residing in their desired path according to
	//first matching term found by SplitOutTheFetchedUrls.
	//Window's library call rename works great to do move.
	//File location is as recorded in the URL LIST note 1.
	//New file path is to be fixed in the URL LIST note 1.
	//Matching is to words recorded in the URL LIST notes.
	//Files listing no matches in the notes are not moved.
	//Each note one looks like this: "#345 34\5mm7it.htm".

	int i;
	for(i = 0; i < nResortedUrlPtrs; i++)
	{
		int PathNo = pResortedUrlPtrs[i].X;
		if(PathNo < NamingListTermsCount)
		{
			//This URL at [i] shows a matching term in its notes.
			//pResortedUrlPtrs[i].pNote points to its note1 item.
			char wk [260];
			char * into = wk;
			char * from = pResortedUrlPtrs[i].pNote + sizeof(char*) + sizeof("#345 ") - 1;
			char * scan = from;
			char * tail = from;
			//fprintf(stdout, "%s; %s;\r\n", NamingListPtrs [PathNo + PathNo + 1], from);
			while(*scan != '\0')
			{
                //Wow! Windows' rename works even across drives.
                //So, I could have a line in URL LIST like this:
                //@a:\bibpath bible
				if(*scan == ':'
				|| *scan == '\\')
				{
					tail = scan + 1;
				}
				scan++;
			}
			*into = '\0';
			//If the original path matches the desired path, do nothing.
			if(tail > from
			&& strnicmp(NamingListPtrs [PathNo + PathNo + 1], from, tail - from) == 0)
                continue;
			//Prepare to rename the file, deal with all errors.
			scan = NamingListPtrs [PathNo + PathNo + 1];
			//Add the new desired "...path\" to the work buffer.
			while((*into=*scan)!='\0' && into < wk + sizeof(wk) - 1)
				scan++, into++;
			//*into = '\0';
			//Append the tail "filexxxx.htm" to the work buffer.
			scan = tail;
			while((*into=*scan)!='\0' && into < wk + sizeof(wk) - 1)
				scan++, into++;
			*into = '\0';
	        if(rename(from, wk) != 0)
			{
				//Write -M errors to stdout, so user can collect them
				fprintf(stdout, "SURF: cannot rename %s to %s: %s.\n", from, wk, strerror(errno));
				continue;
			}
	        //After a successful rename, now have to update note.
            //Let's just assume the note will need a fresh alloc.
			{
			    int ItemSize = sizeof(char*) + sizeof("#345 ") - 1 + strlen(wk) + 1;
				char* Extant = pResortedUrlPtrs[i].pNote;
				char* Item = (char*) malloc(ItemSize);
				if(Item == (char*)0)
				{
					fprintf(stderr, "SURF: Malloc failure changing URLs to move.\n");
	        		exit(1);
		    	}
                //Copy the downstream note-2 pointer
				*(char**)Item = *(char**)Extant;
                //Copy the Quality indicators
	        	strncpy(Item + sizeof(char*), Extant + sizeof(char*), sizeof("#345 ") - 1 );
	        	strcpy(Item + sizeof(char*) + sizeof("#345 ") - 1, wk);
                //Fix the pointer to this note 1. This is the only pointer,
			    //because I nullified the moved item out of original list.
				pResortedUrlPtrs[i].pNote = Item;
                //Free the original note 1
                free (Extant);
			}
		}
	}
	return;
}

void WriteUrlListToScriptFile()
{
	//Output the URLs and notes to LIST and maybe HTM, and free all.
	int i, j, jordinal;
	FILE * fs = fopen(ScriptFilename, "wb");
	if(fs == (FILE*)0)
	{
		fprintf(stderr, "SURF: cannot create %s: %s.\n", ScriptFilename, strerror(errno));
		exit(1);
	}

	//Start with introductory text, and for HTML a nice file header.
	{
		//Get one line of most-used-of-the-most-used-words to adorn files.
		char FrequentWords [120]; //As big as OutputVocabulary's WorkBuffer
		FrequentWords [0] = '\0';
		if(nSortedVocabulary > 0) //I must guarantee nSortedVocabulary > 0
		{
			//Copy here 1 line of the most-used words, about 60 chars,
			//to describe files to annotate the LIST and -W HTML file.
			OutputVocabulary(FrequentWords); //Passing a Non-NULL pointer
			ClearVocabulary();
		}

		//Prefix some introductory text to the URL LIST file.
		fprintf(fs,
		" SURF URL file: '%s'\r\n"
		" %s\r\n"
		"\r\n"
		" # prevents re-downloading web pages that you already have.\r\n"
		" * prevents downloading novel links. Delete * to SURF them.\r\n"
		" You can add @path word... lines to control file locations.\r\n"
		, ScriptFilename, FrequentWords);

		//Prefix an HTML header if also creating that output:
		if(ProgramIsDoingWrite)
		{
			//Make a TITLE: "1234 Surf links: topic topic topic...".
			//Make a H1 line: "1234 Surf links".
			//Make a H2 line: "topic topic topic...".
			//I use binary mode for stdout: then I can put \r\n uniformly.
			fprintf(stdout,
			"<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\r\n"
			"<HTML>\r\n"
			"<HEAD>\r\n"
			"<TITLE>%d SURF links: %s</TITLE>\r\n"
			"</HEAD>\r\n"
			"<BODY>\r\n"
			"<H1>%d SURF links</H1>\r\n"
			"<H2>%s</H2>\r\n"
			" \r\n", nResortedUrlPtrs, FrequentWords, nResortedUrlPtrs, FrequentWords );
		}
	}

	//If there were any, output list of user controls to relocate files.
	//The following is a pretty generic linked list consumer code block.
	if(NamingListHead != (void**)0)
	{
		void** Run = NamingListHead;
		fprintf(fs, "\r\n" );
		do //I proved above that (Run != (void**)0)
		{
			char * TextData = (char*)Run + sizeof(void**);
			fprintf(fs, "@%s\r\n", TextData);
			{
				void** next = (void**)*Run;
				//No, I'm NOT the last user of NamingList... free(Run); 
				Run = next;
			}
		} //do...
		while (Run != (void**)0);
	}

	//This marker helps automate stream edits, and helps user understand.
	fprintf(fs,
		"\r\n\t----LOCAL FILES----\r\n", ScriptFilename );

	//Process the new list to output all the old ('#') links.
	//Outer loop, j, will re-run the new list multiple times
	//to collect by order of the .X keys from the NamingList.
	//URLs whose most-used-words show no matches bear .X max.

	//But first, clone a piece of that loop to prepare local
	//anchors into local anchors in the HTML output version:
	// <UL>
	// <LI><A HREF="#keyword">keyword</A>
	// </UL>
	if(ProgramIsDoingWrite
	&& NamingListTermsCount > 0)
	{
		//Add a list of in-page anchors for the HTML.
        //Process from 0 to NamingListTermsCount - 1.
        //Then do one more loop to do "All The Rest".
		jordinal = 0;
		j = 0;
		fprintf(stdout, "<BR>Selections<UL>\r\n");
		while(j <= NamingListTermsCount)
		{
			//Rather than index J one by one through all the
            //categorization words, group together all those
            //words that point to the same file storage path.
            int jmin = j;
            int jmax = j;
            do
            {
                j++; //Forcing final outer iteration to stop
            } //do...
    		while(j < NamingListTermsCount //protecting next access
                && NamingListPtrs [j+j+1]
                == NamingListPtrs [jmin+jmin+1]);
            jmax = j - 1;

			fprintf(stdout, "<LI><A HREF=\"#");
            if(jmin == NamingListTermsCount)
            {
        	    fprintf(stdout, "AllTheRest");
            }
            else
            {
                int k = jmin;
                do
                {
        			fprintf(stdout, "%s", NamingListPtrs [k+k]);
                } //do...
                while(++k <= jmax);
            }
			fprintf(stdout, "\">(%d)", ++jordinal);
            if(jmin == NamingListTermsCount)
            {
        	    fprintf(stdout, " All The Rest");
            }
            else
            {
                int k = jmin;
                do
                {
        			fprintf(stdout, " %s", NamingListPtrs [k+k]);
                } //do...
                while(++k <= jmax);
            }
			fprintf(stdout, "</A>\r\n");
		}
		fprintf(stdout, "</UL>\r\n");
	}

	//Finally, the promised loop to output fetched URLs.
	j = 0;
    jordinal = 0;
	while(j <= NamingListTermsCount)
	{
        //Process from 0 to NamingListTermsCount - 1.
        //Then do one more loop to do "All The Rest".
		//Rather than index J one by one through all the
        //categorization words, group together all those
        //words that point to the same file storage path.
        int jmin = j;
        int jmax = j;
        do
        {
            j++; //Forcing final outer iteration to stop
        } //do...
    	while(j < NamingListTermsCount //protecting next access
            && NamingListPtrs [j+j+1]
            == NamingListPtrs [jmin+jmin+1]);
        jmax = j - 1;

		if(NamingListTermsCount > 0)
		{
			//If there is a NamingList, show terms in LIST file;
			fprintf(fs, "\r\n [%d]", ++jordinal);
            if(jmin == NamingListTermsCount)
            {
        	    fprintf(fs, " All The Rest");
            }
            else
            {
                int k = jmin;
                do
                {
        			fprintf(fs, " %s", NamingListPtrs [k+k]);
                } //do...
                while(++k <= jmax);
            }
            fprintf(fs, "\r\n\r\n");
	        if(ProgramIsDoingWrite)
            {
        	    //Make in-page anchors for the HTML.
				fprintf(stdout, "\r\n<H2><A NAME=\"");
                if(jmin == NamingListTermsCount)
                {
            	    fprintf(stdout, "AllTheRest");
                }
                else
                {
                    int k = jmin;
                    do
                    {
            			fprintf(stdout, "%s", NamingListPtrs [k+k]);
                    } //do...
                    while(++k <= jmax);
                }
				fprintf(stdout, "\">(%d)", jordinal);
                if(jmin == NamingListTermsCount)
                {
            	    fprintf(stdout, " All The Rest");
                }
                else
                {
                    int k = jmin;
                    do
                    {
            			fprintf(stdout, " %s", NamingListPtrs [k+k]);
                    } //do...
                    while(++k <= jmax);
                }
				fprintf(stdout, "</A></H2>\r\n");
			}
		}
		i = 0;
		while(i < nResortedUrlPtrs)
		{
			//Cloned guts from the loop below.
			//Then I added HTML-to-stdout work.
			//Now the guts are conditional on J.
			if(pResortedUrlPtrs[i].X <= jmax)
			{
				pResortedUrlPtrs[i].X = 12345678; //Never come back here.
				fputs(pResortedUrlPtrs[i].pUrl, fs);
				fputc('\r', fs);
				fputc('\n', fs);
				if(ProgramIsDoingWrite)
				{
					//For HTML, prefix an ordinal to orient user;
					//Output the URL inside an Anchor <A...> tag.
					fprintf(stdout,
					"\r\n<p> %d. <a href =\"%s\">\r\n",
					i + 1, pResortedUrlPtrs[i].pUrl );
				}
				free(pResortedUrlPtrs[i].pUrl); //I last user, so free it.
				//If there are notes for this url,
				//put each one on a new line starting with a space or tab char.
				{
					char * ItemNote = pResortedUrlPtrs[i].pNote;
					int NoteCounter = 0;
					while(ItemNote != (char*)0)
					{
						char * Next = *(char**)ItemNote;
						fputc(' ', fs);
						fputs(ItemNote + sizeof(char*), fs);
						fputc('\r', fs);
						fputc('\n', fs);
						if(ProgramIsDoingWrite)
						{
							NoteCounter++;
							//Skip the first note, such as "#329 32/9tn8it.htm"
							if(NoteCounter != 1)
							{
								fputc(' ', stdout);
								fputs(ItemNote + sizeof(char*), stdout);
								//Add End-Anchor </A> tag after note #2 == title.
								if(NoteCounter == 2)
									fputs("</a><br>", stdout);
								fputc('\r', stdout);
								fputc('\n', stdout);
							}
						}
						free(ItemNote); //I'm last user, so free it.
						ItemNote = Next;				
					}
			   		if(ProgramIsDoingWrite)
			   		{
						//Ensure there is always an End-Anchor </A> tag.
						if(NoteCounter < 2)
							fputs("No Title</a><br>\r\n", stdout);
					}
				}
				fputc('\r', fs);
				fputc('\n', fs);
			}
			i++;
		}
	}

	//I thought I was, but now I AM finished with the NamingList.
	//Free URLS and linked URL notes, then the global array, etc.
	if(NamingListHead != (void**)0)
	{
		void** Run = NamingListHead;
		do //I proved above that (Run != (void**)0)
		{
			void** next = (void**)*Run;
			free(Run); //I'm last user of NamingList, emptying it.
			Run = next;
		} //do...
		while (Run != (void**)0);
		//Keep empty list's pointer integrity, in case resurrected.
		NamingListTail = (void**) & NamingListHead;
		NamingListHead = (void**) 0;
		//That list implied 2 other data stores, now free them too:
		if(NamingListMatchTerms != (char*)0)
		{
			free(NamingListMatchTerms);
			NamingListMatchTerms = (char*)0;
		}
		NamingListWorstCaseSize = 0;
		if(NamingListPtrs != (char**)0)
		{
			free(NamingListPtrs);
			NamingListPtrs = (char**)0;
		}
		NamingListTermsCount = 0;
	}

	//I am finished with the resorting vector.
	free(pResortedUrlPtrs); //Created recently in SplitOutTheFetchedUrls.
	pResortedUrlPtrs = (struct UrlPtrsX *) 0; //Not that anyone ever looks.
	nResortedUrlPtrs = 0;

	//This marker helps automate stream edits, and helps user understand.
	fprintf(fs,	"\t----NOVEL LINKS----\r\n");

	//And I can wrap up the HTML listing output now.
	if(ProgramIsDoingWrite)
	{
		//Enabling Windows 95 accessibility option "high contrast" makes
		//status line fat; This <P> helps Netscape not hide final text.
		fprintf(stdout,
		"\r\n"
		"<P>\r\n"
		"</HTML>\r\n"
		"</BODY>\r\n" );
	}

	//Reprocess original list to output all the new ('*') links.
	i = 0;
	while(i < nSortedUrlPtrs)
	{
		//Skip over any items that were nulled out of this list.
		if(pSortedUrlPtrs[i].pUrl != (char*)0)
		{
			fputs(pSortedUrlPtrs[i].pUrl, fs);
			fputc('\r', fs);
			fputc('\n', fs);
			free(pSortedUrlPtrs[i].pUrl);
			//If there are notes for this url,
			//put each one on a new line starting with a space or tab char.
			{
				char * ItemNote = pSortedUrlPtrs[i].pNote;
				while(ItemNote != (char*)0)
				{
					char * Next = *(char**)ItemNote;
					fputc(' ', fs);
					fputs(ItemNote + sizeof(char*), fs);
					fputc('\r', fs);
					fputc('\n', fs);
					free(ItemNote);
					ItemNote = Next;				
				}
			}
			fputc('\r', fs);
			fputc('\n', fs);
		}
		i++;
	}
	if(fclose(fs) == -1)
	{
		fprintf(stderr, "SURF: error closing file %s: %s.\n", ScriptFilename, strerror(errno));
		exit(1);
	}
	//Restore the empty condition of the list
	if(pSortedUrlPtrs != (struct UrlPtrs *)0)
		free(pSortedUrlPtrs);
	pSortedUrlPtrs = (struct UrlPtrs *)0;
	nMallocUrlPtrs = 0;
	nSortedUrlPtrs = 0;
	return;
}

void FetchObjectFromUrlToFile(char * pUrl)
{
	//This routine started from the MSVC++ 5.0 sample code program "TEAR".
	CInternetSession session(SurfVersion, 1, PRE_CONFIG_INTERNET_ACCESS);
	CHttpConnection* pServer = NULL;
	CHttpFile* pFile = NULL;
	static char UniqueFilename [] = "012345.htm";
	char UrlText [500]; //Arbitrary limit on overall URL size
	{
		time_t ltime;
		time( &ltime );
		do {
			unsigned int u = (unsigned int) ltime;
			static char NiceLetters [] = "23456789abcdefghijklmnoprstuvxyz";
			UniqueFilename [0] = NiceLetters [u & 31]; u >>= 5;
			UniqueFilename [1] = NiceLetters [u & 31]; u >>= 5;
			UniqueFilename [2] = NiceLetters [u & 31]; u >>= 5;
			UniqueFilename [3] = NiceLetters [u & 31]; u >>= 5;
			UniqueFilename [4] = NiceLetters [u & 31]; u >>= 5;
			UniqueFilename [5] = NiceLetters [u & 31];
			ltime++;
		} while(_access(UniqueFilename, 00) == 0);
		LocalFilename = UniqueFilename;
	}
	ObjectFile = fopen( LocalFilename, "w");
	//Give a terse indication of progress in urls and filenames:
	fprintf(stderr, "%s -> %s\n", pUrl, LocalFilename);
	if(ObjectFile == (FILE*)0)
	{
		fprintf(stderr, "SURF: cannot create %s: %s.\n", LocalFilename, strerror(errno));
		exit(1);
	}

	//Canonicalize this Url and save as Base Url to resolve local Anchor Urls.
	//Reset Base Url so that ordinary anchor URL parsing process will suffice.
	memset( (void*) & BaseUrlParts, 0, sizeof(BaseUrlParts));

	//Because GetLinkUrlParts is destructive, make a copy of URL to fetch:
	strcpy(inbuf, pUrl);				//First misuse of inbuf

	if(GetLinkUrlParts(inbuf, inbuf + strlen(inbuf)) == 0)
	{
		//Reconstitute a perfect Url.
		char *from, *into = inbuf;	  //Second misuse of inbuf
		from = LinkUrlParts.scheme;
		if((*into=*from)!='\0') do from++, into++; while((*into=*from)!='\0');
		*into++ = ':';
		*into++ = '/';
		*into++ = '/';
		from = LinkUrlParts.netlocn;
		if((*into=*from)!='\0') do from++, into++; while((*into=*from)!='\0');

		//Often, URL's urlpath may be empty, having only a net location.
		//I wonder if all servers would match "domain.x/" to "domain.x".

		//The reason I worry, is that Netscape Navigator will not resolve
		//relative urls correctly when browsing these locally saved files
		//as annotated with BASE HREF="domain.x" held on local hard drive.

		//I'm going to try appending a '/' to URLS that are just a domain name.
		//But this 3/15/98 change, 2 places, might interfere with url fetching.
		if(*LinkUrlParts.urlpath == '\0'
		&& *LinkUrlParts.parameter == '\0'
		&& *LinkUrlParts.query == '\0')
		{
			*into++ = '/';
			*into = '\0';
		}
		else
		{
			from = LinkUrlParts.urlpath;
			if((*into=*from)!='\0') do from++, into++; while((*into=*from)!='\0');
			from = LinkUrlParts.parameter;
			if((*into=*from)!='\0') do from++, into++; while((*into=*from)!='\0');
			from = LinkUrlParts.query;
			if((*into=*from)!='\0') do from++, into++; while((*into=*from)!='\0');
		}
		if(into - inbuf >= sizeof(UrlText))
		{
			fprintf(ObjectFile, "Error: Unsuitable Url: \"%s\".\n", pUrl);
			goto MyCloseFileBeforeReturn;
		}
		strcpy(UrlText, inbuf);		 //Done with misuse of inbuf
		fprintf(ObjectFile, "<BASE HREF=%s>\n", UrlText);

		//This BaseUrlParts will be used to reconstitute incomplete Anchor Urls:
		memcpy( (void*) & BaseUrlParts, (void*) & LinkUrlParts, sizeof(BaseUrlParts));
	}
	else
	{
		fprintf(ObjectFile, "Error: Unsuitable Url: \"%s\".\n", pUrl);
		goto MyCloseFileBeforeReturn;
	}

	try
	{
		CString strServerName;
		CString strObject;
		INTERNET_PORT nPort;
		DWORD dwServiceType;
		if(AfxParseURL(pUrl, dwServiceType, strServerName, strObject, nPort))
		{
			//Currently, I only support the http: method.
			if(dwServiceType == INTERNET_SERVICE_HTTP)
			{
				pServer = session.GetHttpConnection(strServerName, nPort);
				pFile = pServer->OpenRequest(CHttpConnection::HTTP_VERB_GET,
					strObject, NULL, 1, NULL, NULL, INTERNET_FLAG_EXISTING_CONNECT);
				pFile->AddRequestHeaders(_T("Accept: text/*\r\nUser-Agent: SURF\r\n"));
				pFile->SendRequest();
				//Clone and adapt the successful file parsing loop...
				{
					//One of my favorite file input methods is to read whole
					//next disk block just past the tail of unprocessed data.
					int noffset = 0;
					for(;;)
					{
						int nread;
						//Adapt my loop to the stream read of the sample code.

						char *cp = pFile->ReadString((inbuf+4) + noffset, 4096);
						//EOF gives null ptr.
						if(cp == (char*)0)
							nread = 0;
						else
							nread = strlen(cp);
						if(nread <= 0)
						{
							if(noffset == 0)
							{
								//At EOF, and nothing was left from prior loop.
								break;
							}
							//Tail data from prior loop lacks a space at EOF.
							//Append a final space character to assist parse.
							(inbuf+4) [noffset] = ' ';
							noffset++;
						}
						//This may copy back some tail of data partly processed.
						noffset = MultiPassInbufConversion(noffset + nread);
						if(noffset == -1)
							break;  //unacceptable garbage or binary data
						if(nread == 0)
							break;  //Input had reached EOF
					} //for...
				}
				pFile->Close();
				pServer->Close();
			}
		}
	}
	catch (CInternetException* pEx)
	{
		// catch errors from WinINet
		TCHAR szErr[1024];
		pEx->GetErrorMessage(szErr, 1024);
		fprintf(stderr, "Error: (%d) %s.\n", pEx->m_dwError, (LPCTSTR)szErr);
        {
            //I should get all words of the error message, but get this:
            char ISaid [] = "Error";
            AddWordToVocabulary(ISaid, ISaid + sizeof(ISaid) - 1);
        }
		fprintf(ObjectFile, "\nError: (%d) %s.\n", pEx->m_dwError, (LPCTSTR)szErr);
		pEx->Delete();
	}

	//Flush any internal parse buffers, also restore all states.
	ProcessForEndOfInputFile();

MyCloseFileBeforeReturn:;

	//These were already created before that goto too.
	if (pFile != NULL)
		delete pFile;
	if (pServer != NULL)
		delete pServer;
	session.Close();
	if(fclose(ObjectFile) == -1)
	{
		fprintf(stderr, "SURF: error closing file %s: %s.\n", LocalFilename, strerror(errno));
		exit(1);
	}
	ObjectFile = (FILE*)0;

	//Rename to alternate filename according to "@path word" controls.
	{
        //This requires cloning the loop to look for match word,
        //and the work of the MoveAboutTheFetchedUrls() routine.

        ///wip... This would be the note with old path to fix:
	    ///char *Item = pSortedUrlPtrs[LastUrlPtrSlot].pNote;
	}
	return;
}

int RandomlyFetchOneURLFromUrlList()
{
	//Choose one URL at random, fetch that Internet object, mark URL.
	//Return -1 to stop caller's loop when URL list is exhausted.
	//The 2/98 version didn't fetch randomly when had over 32K links:
	int i = (rand() * rand()) % nSortedUrlPtrs;
	int n = i;
	//Search for the next URL that was not fetched
	for(;;) {
		//Fall out of loop when i locates an unfetched URL to fetch.
		//Do not fetch if "*" is the first character of the first note.
		//Nor if "#" is the first character of the first note.
		if(pSortedUrlPtrs[i].pNote == (char*)0
		|| *(pSortedUrlPtrs[i].pNote + sizeof(char*)) != '*'
		&& *(pSortedUrlPtrs[i].pNote + sizeof(char*)) != '#')
			break; //That URL at random i was acceptable
		if(++i == nSortedUrlPtrs)
			i = 0; //Keep looking: increment and wrap around
		if(i == n) //Back to the original random choice.
			return -1; //signal no more URLs to fetch with -1
	}
	//Before this file is fetched, mark out this slot [i] as fetched.
	//This is necessary for files without "nice" urls, but generally
	//ProcessForEndOfInputFile will replace with a note starting '#',
	//and I do ProcessForEndOfInputFile via FetchObjectFromUrlToFile.
	{
		static char Blurb [] = "#(FETCHED)";
		char* Extant = pSortedUrlPtrs[i].pNote;
		char* Item = (char*) malloc(sizeof(char*) + sizeof(Blurb));
		if(Item == (char*)0)
		{
			fprintf(stderr, "SURF: Malloc failure adding URLs to list.\n");
			exit(1);
		}
		*(char**)Item = Extant;
		strcpy(Item + sizeof(char*), Blurb);
		pSortedUrlPtrs[i].pNote = Item;
	}

	FetchObjectFromUrlToFile(pSortedUrlPtrs[i].pUrl);

	//Output fetched page's analysis, so user can gauge productivity.
	//Notice that my i has become obsolete, if some links were added.
	//ProcessForEndOfInputFile will write a first note with '#' mark,
	//and I do ProcessForEndOfInputFile via FetchObjectFromUrlToFile.
	if(LastUrlPtrSlot != -1)
	{
		char * Item = pSortedUrlPtrs[LastUrlPtrSlot].pNote;
		while(Item != (char*)0)
		{
			char * Next = *(char**)Item;
	        fprintf(stdout, " %s\n", Item + sizeof(char*));
			Item = Next;
		}
	}
	return 0;
}

void PrepareFileSavingControls()
{
	//Call here for -F and -M command line switches.
	//No, everybody needs it. But only -F, -M mkdir.
	//I have already done ReadScriptFileToUrlList(),
	//and any lines in LIST starting with @ specify
	//paths to store files containing a listed term
	//in the most used words list.
	//Caller guaranteed NamingListWorstCaseSize > 0.
	void** Run;
	char *fill, *TopOfRecord;
	//I loosely counted the bytes needed when saving these lines.
	//Hold paths and words as a wad of contiguous asciiz strings.
	NamingListMatchTerms = (char*)malloc(NamingListWorstCaseSize);
	if(NamingListMatchTerms == (char*)0)
	{
		fprintf(stderr, "SURF: Malloc failure making file saving list.\n");
		exit(1);
	}
	//I use, and run through items in NamingList, but do not empty it.
	Run = NamingListHead;
	fill = NamingListMatchTerms;
	if(Run != (void**)0)
	do
	{
		char * TextData = (char*)Run + sizeof(void**);
		TopOfRecord = fill;
		//fprintf(stdout, "Filenaming control: <%s>\r\n", TextData);
		//first skip any space or tab chars.
		while(*TextData == ' ' || *TextData == '\t')
			TextData++;
		//second, process all non-space chars as a file path.
		for(;;)
		{
			//Stop at every path separator character;
			//Make that directory if it doesn't exist.
			//Oh, I must call this at end of path too.
			*fill = *TextData;
			if(*TextData == '/' || *TextData == '\\'
			|| *TextData == ' ' || *TextData == '\t' || *TextData == '\0')
			{
				*fill = '\0'; //asciz terminate the partial path
				if(*TopOfRecord != '\0'
				&& _access(TopOfRecord, 00) != 0)
				{
					//fprintf(stdout, "Need directory %s.\n", TopOfRecord);
					//If it does not exist, try to make directory.
	                //Gate this for only -F and -M command lines.
                    if(ProgramIsDoingMkdir)
                    {
						if(_mkdir(TopOfRecord) == -1)
						{
							fprintf(stderr, "SURF: cannot make directory %s: %s.\n", TopOfRecord, strerror(errno));
							exit(1);
						}
				    }
				}
				else
				{
					//Strictly, I ought to check if file *IS* a directory.
					//fprintf(stdout, "Have directory %s.\n", TopOfRecord);
				}
				*fill = '\\'; //regulate the path separator char
			}
			fill++;
			if(*TextData == ' ' || *TextData == '\t' || *TextData == '\0')
				break;
			TextData++;
		}
		*fill++ = '\0'; //asciz terminate the overall path
		//Wierd loop construction above guaranteed \ at end.
		//process rest of record copying and counting terms.
		for(;;)
		{
			//This loop does not guarantee terms after path.
			if(*TextData == '\0')
				break;
			//skip whitespace
			while(*TextData == ' ' || *TextData == '\t')
				TextData++;
			if(*TextData == '\0')
				break;
			//collect printables
			while(*TextData != ' ' && *TextData != '\t' && *TextData != '\0')
			{
				*fill = *TextData; //copy term
				fill++;
				TextData++;
			}
			NamingListTermsCount++;
			*fill++ = '\0'; //asciz terminate the term
		}
		*fill++ = '\0'; //double-null terminate each record
		//Advance from this record (still NOT freed!) to next record.
		Run = (void**)*Run;
	} //do...
	while (Run != (void**)0);
	*fill++ = '\0'; //triple-null end of list in case I want to use it.
	//I need memory to set up two pointers for each match term.
	if(NamingListTermsCount != 0)
	{
		NamingListPtrs = (char**)malloc(NamingListTermsCount * 2 * sizeof(char*));
		if(NamingListPtrs == (char**)0)
		{
			fprintf(stderr, "SURF: Malloc failure making file saving list.\n");
			exit(1);
		}
	}
	//Rescan the Wad o' Text to hook up pointers to term and path.
	{
		int i = 0;
		char * scan = NamingListMatchTerms;
		//Process records until the triple null.
		while(*scan != 0)
		{
			TopOfRecord = scan;
			//skip over the directory path string
			while(*scan != 0)
				scan++;
			scan++;
			//Now process any asciz words found
			while(*scan != 0)
			{
				//every pair of pointers will locate one term, then a path.
				NamingListPtrs [i++] = scan; //one word
				NamingListPtrs [i++] = TopOfRecord; //then its path
				//skip over the word
				while(*scan != 0)
					scan++;
				scan++;
			}
			scan++;
		}
		//pruune auuthor fauults
		if(i != 2 * NamingListTermsCount)
		{
			fprintf(stderr, "SURF: Program error #1.\r\n");
			exit(1);
		}
	}
	//Show me...
	//{ int i; for(i=0; i < NamingListTermsCount; i++) fprintf(stdout, "<%s>=<%s>\r\n", NamingListPtrs [i+i], NamingListPtrs [i+i+1]); }
	return;
}

int main(int argc, char* argv [])
{
	//Using MFC API for Internet access forces me to do things like this.
	//As noted, compiler Settings: C/C++, Code Generation: Multithreaded.
	if (!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0))
	{
		fprintf(stderr, "MFC Failed to initialize.\n");
		exit(1);
	}

#if 0
	//Enable these selfchecks once after table edits.
	if (SortedTagNames [mu_lowsentinel] [0] != '!'
	||  SortedTagNames [mu_topsentinel] [0] != '~'
	||  MarkupTagTypes [mu_lowsentinel] != -1
	||  MarkupTagTypes [mu_topsentinel] != -1
	||  (sizeof(SortedEntityNames) & 7) != 1
	||  (sizeof(CommonWordList) & 15) != 1)
	{
		fprintf(stderr, "Bad Tables.\n");
		exit(1);
	}
	{
		//More selfchecks on tables.
		int i;
		for(i=0; i < sizeof(SortedTagNames) / sizeof(*SortedTagNames) - 1; i ++)
			if(strcmp(SortedTagNames [i], SortedTagNames [i+1]) >= 0)
			{
				fprintf(stderr, "Bad TagNames %d.\n", i);
				exit(1);
			}
		for(i=0; i < sizeof(SortedEntityNames) - 1 - 8; i += 8)
			if(strcmp(SortedEntityNames + i, SortedEntityNames + i + 8) >= 0)
			{
				fprintf(stderr, "Bad EntNames %d.\n", i);
				exit(1);
			}
		for(i=0; i < sizeof(CommonWordList) - 1 - 16; i += 16)
			if(strcmp(CommonWordList + i, CommonWordList + i + 16) >= 0)
			{
				fprintf(stderr, "Bad ComnWds %d.\n", i);
				exit(1);
			}
	}
#endif //if 0

	//Analyze the command line.
	if(argc < 2)
	{
		// argc == 1 means no arguments: "SURF"
		//I require at least one LIST filename as first argument.
		GiveUsageMessage();
		exit(1);
	}

	//Skip this for argc == 2 means only a list argument: "SURF LIST"
	if(argc > 2)
	{
		// argc == 3 means one list and one flag argument: "SURF LIST -X"
		//Check the overall invocation syntax:
		//Any second argument must be one -FLAG: {a, f, m, q, w}.
		if(argv [2] [0] != '-')
		{
			fprintf(stderr, "SURF: Second argument must be (-a, -f, -m, -q, or -w).\n");
			GiveUsageMessage();
			exit(1);
		}
		CmdLineLetter = argv [2] [1] | ' '; // lowercase
		//All flags are just one hyphen and one character.
		if(argv [2] [2] != '\0')
			CmdLineLetter = '?'; //Invalidate
		
		if(CmdLineLetter != 'a'
		&& CmdLineLetter != 'f'
		&& CmdLineLetter != 'm'
		&& CmdLineLetter != 'q'
		&& CmdLineLetter != 'w')
		{
			fprintf(stderr, "SURF: Flag (%s) not understood.\n", argv [2]);
			GiveUsageMessage();
			exit(1);
		}
	}
	{
		//Try to access ScriptFilename, supplying no filespec expansion.
		//This must be an exact filename for the LIST file.
		ScriptFilename = argv [1];
		if(_access(ScriptFilename, 00) == 0)
		{
			//If it exists, I import it now, and will later re-write update.
			ReadScriptFileToUrlList();
			//Now, lines starting with '@' signify something special to me.
			if(CmdLineLetter == 'f'
			|| CmdLineLetter == 'm')
                ProgramIsDoingMkdir = 1; //For the -F and -M flags
			if(NamingListWorstCaseSize > 0)
				PrepareFileSavingControls();
		}
		else
		{
			//If it doesn't exist, make sure right away I can create the file.
			FILE * fs = fopen(ScriptFilename, "wb");
			if(fs == (FILE*)0
			|| fclose(fs) == -1)
			{
				fprintf(stderr, "SURF: cannot create list file %s: %s.\n", ScriptFilename, strerror(errno));
				exit(1);
			}
		}
	}



	switch(CmdLineLetter)
	{

	//Default value is ' ' when no -FLAGS are given
	case ' ': //Surf the list file

		if(nSortedUrlPtrs == 0)
		{
			fprintf(stderr, "SURF: There were no URLS to fetch in %s.\n", ScriptFilename);
			exit(1);
		}
		//I select the URLs to fetch in a random order,
		//to avoid being too burdensome a net citizen.
		//Seed randoms from the system clock.
		srand((unsigned)time((time_t*)0));
		//Reset these when writing to a local file.
		DiscardingTextNow = DiscardingAllText = 0;

		for(;;)
		{
			if(RandomlyFetchOneURLFromUrlList() == -1)
			{
				fprintf(stderr, "SURF: No more URLS to fetch in %s.\n", ScriptFilename);
				fprintf(stderr, "SURF: You must remove the '*' after URLs you want to fetch next.\n");
				break;
			}
			if(_kbhit() != 0)
			{
				fprintf(stderr, "SURF: Stopping due to your keystroke.\n");
				break;
			}
		}

		//These two calls must be used as a pair.
		SplitOutTheFetchedUrls();
		WriteUrlListToScriptFile();
		break;

	case 'a': //Augment list with local FILE1, FILE2,...

		{
			int ai = 2; //Skip argv[1]:LIST and argv[2]:-FLAG
			while (++ai < argc)
			{
				//If filespec finds any directory,
				//I'll analyze all files under it.
				RecursivePathExpansion(0, argv [ai]);
			}
		}
		//Process the file list, and free all mallocs.
		if (FileListHead != (void**)0)
		do
		{
			char * InputFilename = LocalFilename = (char*)FileListHead + sizeof(void**);
			int hfi = _open(InputFilename, _O_BINARY|_O_RDONLY, _S_IREAD|_S_IWRITE);
			if(hfi == -1)
			{
				fprintf(stderr, "SURF: cannot open to read file %s: %s.\n", InputFilename, strerror(errno));
				//But do not exit program for that.
			}
			else
			{
				//One of my favorite file input methods is to read whole
				//next disk block just past the tail of unprocessed data.
				int noffset = 0;
				//Give a terse indication of progress in filenames:
				fprintf(stderr, "%s\n", InputFilename);
				for(;;)
				{
					int nread = _read(hfi, (inbuf+4) + noffset, 4096);
					if(nread <= 0)
					{
						//EOF gives 0; Errors -1.
						if(nread < 0)
						{
							fprintf(stderr, "SURF: error reading file %s: %s.\n", InputFilename, strerror(errno));
							exit(1);
						}
						if(noffset == 0)
						{
							//At EOF, and nothing was left from prior loop.
							break;
						}
						//Tail data from prior loop lacks a space at EOF.
						//Append a final space character to assist parse.
						(inbuf+4) [noffset] = ' ';
						noffset++;
					}
					//This may copy back some tail of data partly processed.
					noffset = MultiPassInbufConversion(noffset + nread);
					if(noffset == -1)
						break;  //unacceptable garbage or binary data
					if(nread == 0)
						break;  //Input had reached EOF
				} //for...
				//Flush any internal parse buffers, also restore all states.
				//This case parsing local files did not use FetchObjectFromUrlToFile,
				//so I am the only case to itself call ProcessForEndOfInputFile.
				ProcessForEndOfInputFile();

				if(_close(hfi) == -1)
				{
					fprintf(stderr, "SURF: error closing file %s: %s.\n", InputFilename, strerror(errno));
					exit(1);
				}
			}
			{
				void** next = (void**)*FileListHead;
				free(FileListHead);
				FileListHead = next;
			}
			if(_kbhit() != 0)
			{
				fprintf(stderr, "SURF: Stopping due to your keystroke.\n");
				break;
			}
		} //do...
		while (FileListHead != (void**)0);
		//These two calls must be used as a pair.
		SplitOutTheFetchedUrls();
		WriteUrlListToScriptFile();
		break;

	case 'f': //Fetch web pages from URL1, URL2...

		ProgramIsDoingFetch = 1; //For the -F flag

		//I use binary mode for stdout: then I can put \r\n uniformly.
		_setmode(_fileno(stdout), _O_BINARY);

		{
			int ai = 2; //Skip argv[1]:LIST and argv[2]:-FLAG
			while (++ai < argc)
			{
				//Reset these when writing to a local file.
				DiscardingTextNow = DiscardingAllText = 0;
				//Fetch the single Internet object.
				if(strnicmp(argv [ai], "http://", 7) == 0)
				{
					//The user was not a lazy typist.
					FetchObjectFromUrlToFile(argv [ai]);
				}
				else
				{
					//Allow that user need not type in "http://" atop url.
					//Remember, right now surf only processes HTTP method.
					char WithHttp [300]; //This is no limit if you type it.
					strcpy (WithHttp, "http://");
					strncpy (WithHttp + 7, argv[ai], sizeof(WithHttp) - 1 - 7);
					//Curiously, strncpy doesn't guaranteed a final \0.
					WithHttp [sizeof(WithHttp) - 1] = '\0';
					FetchObjectFromUrlToFile(WithHttp);
				}
				//Flush any internal parse buffers, also restore all states.
				//But, I do ProcessForEndOfInputFile via FetchObjectFromUrlToFile.
				if(_kbhit() != 0)
				{
					fprintf(stderr, "SURF: Stopping due to your keystroke.\n");
					break;
				}
			}
		}
		//These two calls must be used as a pair.
		SplitOutTheFetchedUrls();
		WriteUrlListToScriptFile();
		break;

	case 'q': //Add search engine queries to list.

		{
			//First, let's assemble the query terms with "+" for spaces.
			char * into = QueryTopicString;
			int ai = 2; //Skip argv[1]:LIST and argv[2]:-FLAG
			while (++ai < argc)
			{
				char * from = argv [ai];
				while((*into=*from) != '\0'
				&& into < QueryTopicString + sizeof(QueryTopicString) - 1)
				{
					//Convert embedded spaces into plus signs:
					if(*from == ' ')
						*into = '+';
					from++, into++; //Copy one search query word
				}
				//Separate query words with + instead of whitespace.
				if(into < QueryTopicString + sizeof(QueryTopicString) - 1)
					*into++ = '+';
			}
			if(into > QueryTopicString)
	   			into [-1] = '\0'; //Discard final "+", which IS present.
			else
				break;
		}
		{
			//Now let's run through the predefined search engine queries.
			//Each two lines gets added as a query and as a note to LIST.
			int qi = 0;
			for(;;)
			{
				char JoinedSearchUrl [500];
				strcpy(JoinedSearchUrl, SearchEngineQueriesAndTitles [qi]);
				strcat(JoinedSearchUrl, QueryTopicString);

				//Canonicalize and process absolute or (base+relative) URLs.
				if(GetLinkUrlParts(JoinedSearchUrl, JoinedSearchUrl + strlen(JoinedSearchUrl) ) == 0
				&& CombineLinkUrlWithBase() == 0) //Don't say && LinkUrlPartsIsAcceptable())
				{
					AddLinkUrlPartsToUrlList(); //URL from SearchEngineQ... plus QueryTopics.
					MergeNoteAtLastUrlPtrSlot(SearchEngineQueriesAndTitles [qi+1]);
				}
				if(_kbhit() != 0)
				{
					fprintf(stderr, "SURF: Stopping due to your keystroke.\n");
					break;
				}
				qi += 2;
				if(qi >= sizeof(SearchEngineQueriesAndTitles)/sizeof(*SearchEngineQueriesAndTitles))
					break; //do not exceed max array index (if well paired)
			}
		}
		//These two calls must be used as a pair.
		SplitOutTheFetchedUrls();
		WriteUrlListToScriptFile();
		break;

	case 'm': //Move files around according to most-used words.

		//These two calls (Split and Write) must be used as a pair.
		//The re-categorized fetched URL lists will control (Move).
		SplitOutTheFetchedUrls();
		MoveAboutTheFetchedUrls();
		WriteUrlListToScriptFile();
		break;

	case 'w': //Write an HTML file of fetched links

		ProgramIsDoingWrite = 1; //For the -W flag

		//I use binary mode for stdout: then I can put \r\n uniformly.
		_setmode(_fileno(stdout), _O_BINARY);

		SplitOutTheFetchedUrls();
		WriteUrlListToScriptFile();

		//If I was religious today, I'd close it and check for errors.
		if(fclose(stdout) == -1)
		{
			fprintf(stderr, "SURF: error closing stdout: %s.\n", strerror(errno));
			exit(1);
		}
		break;
	}
	return 0; // exit(0)
}
